-- originally created by clunatic
MeleeHumanBehaviors = {}

function MeleeHumanBehaviors.SetShootingSkill(AI)
	local skill = 50
	local Activ = ActivityMan:GetActivity()
	if Activ then
		skill = Clamp(Activ.Difficulty or 50, 100, 0)
	end
	
	-- player AI skill increase when the difficulty is below "normal", from 50 at "normal" to 100 at "cake"
	if AI.isPlayerOwned then
		skill = Clamp(100 * math.exp(-0.01458*skill), 100, 50)
	else -- enemy AI skill increase with difficulty, from 23 at "cake" to 100 at "nuts!"
		skill = Clamp( 23 * math.exp( 0.01458*skill), 100, 0)
	end
	
	if skill < 50 then	-- the AI shoot later and tracks the target slower
		AI.aimSpeed = -0.028 * skill + 3.2 -- affects the delay before the shooting starts [3.2 .. 1.8]
		AI.aimSkill = -0.012 * skill + 2.0 -- affects the precision of the shots [2.0 .. 1.4]
	else	-- the AI shoot sooner and with slightly better precision
		AI.aimSpeed = -0.026 * skill + 3.1 -- [1.8 .. 0.5]
		AI.aimSkill = -0.012 * skill + 2.0 -- [1.4 .. 0.8]
	end
end

-- spot targets by casting a ray in a random direction
function MeleeHumanBehaviors.LookForTargets(AI, Owner)
	local viewAngDeg = RangeRand(35, 85)
	if AI.deviceState == AHuman.AIMING then
		viewAngDeg = 20
	end
	
	local FoundMO = Owner:LookForMOs(viewAngDeg, rte.grassID, false)
	if FoundMO then
		local HitPoint = SceneMan:GetLastRayHitPos()
		if not AI.isPlayerOwned or not SceneMan:IsUnseen(HitPoint.X, HitPoint.Y, Owner.Team)
			or not SceneMan:IsUnseen(FoundMO.Pos.X, FoundMO.Pos.Y, Owner.Team)
		then	-- AI-teams ignore the fog
			return FoundMO, HitPoint
		end
	end
end

-- brains spot targets by casting rays at all neraby enemy actors
function MeleeHumanBehaviors.CheckEnemyLOS(AI, Owner)
	if not AI.Enemies then	-- add all enemy actors on our screen to a table and check LOS to them, one per frame
		AI.Enemies = {}
		for Act in MovableMan.Actors do
			if Act.Team ~= Owner.Team then
				if not AI.isPlayerOwned or not SceneMan:IsUnseen(Act.Pos.X, Act.Pos.Y, Owner.Team) then	-- AI-teams ignore the fog
					local Dist = SceneMan:ShortestDistance(Owner.ViewPoint, Act.Pos, false)
					if (math.abs(Dist.X) - Act.Diameter < FrameMan.PlayerScreenWidth * 0.6) and
						(math.abs(Dist.Y) - Act.Diameter < FrameMan.PlayerScreenHeight * 0.6)
					then
						table.insert(AI.Enemies, Act)
					end
				end
			end
		end
		
		return MeleeHumanBehaviors.LookForTargets(AI, Owner)	-- cast rays like normal actors occasionally
	else
		local Enemy = table.remove(AI.Enemies)
		if Enemy then
			if MovableMan:ValidMO(Enemy) then
				local Origin
				if Owner.EquippedItem and AI.deviceState == AHuman.AIMING then
					Origin = Owner.EquippedItem.Pos
				else
					Origin = Owner.EyePos
				end
				
				local LookTarget
				if Enemy.ClassName == "ADoor" then
					local Door = ToADoor(Enemy).Door
					if Door and Door:IsAttached() then
						LookTarget = Door.Pos
					else
						return MeleeHumanBehaviors.LookForTargets(AI, Owner)		-- this door is destroyed, cast rays like normal actors
					end
				else
					LookTarget = Enemy.Pos
				end
				
				-- cast at body
				if not AI.isPlayerOwned or not SceneMan:IsUnseen(LookTarget.X, LookTarget.Y, Owner.Team) then	-- AI-teams ignore the fog
					local Dist = SceneMan:ShortestDistance(Owner.ViewPoint, LookTarget, false)
					if (math.abs(Dist.X) - Enemy.Radius < FrameMan.PlayerScreenWidth * 0.52) and
						(math.abs(Dist.Y) - Enemy.Radius < FrameMan.PlayerScreenHeight * 0.52)
					then
						local Trace = SceneMan:ShortestDistance(Origin, LookTarget, false)
						local ID = SceneMan:CastMORay(Origin, Trace, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, false, 5)
						if ID < rte.NoMOID then
							local MO = MovableMan:GetMOFromID(ID)
							if MO and ID ~= MO.RootID then
								MO = MovableMan:GetMOFromID(MO.RootID)
							end
							
							return MO, SceneMan:GetLastRayHitPos()
						end
					end
				end
				
				-- no LOS to the body, cast at head
				if Enemy.EyePos and (not AI.isPlayerOwned or not SceneMan:IsUnseen(Enemy.EyePos.X, Enemy.EyePos.Y, Owner.Team)) then	-- AI-teams ignore the fog
					local Dist = SceneMan:ShortestDistance(Owner.ViewPoint, Enemy.EyePos, false)
					if (math.abs(Dist.X) < FrameMan.PlayerScreenWidth * 0.52) and
						(math.abs(Dist.Y) < FrameMan.PlayerScreenHeight * 0.52)
					then
						local Trace = SceneMan:ShortestDistance(Origin, Enemy.EyePos, false)
						local ID = SceneMan:CastMORay(Origin, Trace, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, false, 5)
						if ID < rte.NoMOID then
							local MO = MovableMan:GetMOFromID(ID)
							if MO and ID ~= MO.RootID then
								MO = MovableMan:GetMOFromID(MO.RootID)
							end
							
							return MO, SceneMan:GetLastRayHitPos()
						end
					end
				end
			end
		else
			AI.Enemies = nil
			return MeleeHumanBehaviors.LookForTargets(AI, Owner)		-- cast rays like normal actors occasionally
		end
	end
end

function MeleeHumanBehaviors.CalculateThreatLevel(MO, Owner)
	-- prioritize closer targets
	local priority = -SceneMan:ShortestDistance(Owner.Pos, MO.Pos, false).Largest / FrameMan.PlayerScreenWidth
	
	-- prioritize the weaker humans over crabs
	if MO.ClassName == "AHuman" then
		if MO.FirearmIsReady then	-- prioritize armed targets
			priority = priority + 1.0
		else
			priority = priority + 0.5
		end
	elseif MO.ClassName == "ACrab" then
		if MO.FirearmIsReady then	-- prioritize armed targets
			priority = priority + 0.7
		else
			priority = priority + 0.3
		end
	end
	
	return priority - MO.Health / 500	-- prioritize damaged targets
end

function MeleeHumanBehaviors.ProcessAlarmEvent(AI, Owner)
	AI.AlarmPos = nil
	
	local loudness, AlarmVec
	local canSupress = not AI.flying and Owner.FirearmIsReady and Owner.EquippedItem:HasObjectInGroup("Explosive Weapons")
	for Event in MovableMan.AlarmEvents do
		if Event.Team ~= Owner.Team then	-- caused by some other team's activites - alarming!
			loudness = 30 + FrameMan.PlayerScreenWidth * 0.6 * Owner.Perceptiveness * Event.Range	-- adjust the audiable range to the screen resolution
			AlarmVec = SceneMan:ShortestDistance(Owner.EyePos, Event.ScenePos, false)	-- see how far away the alarm situation is
			if AlarmVec.Largest < loudness then	-- only react if the alarm is within hearing range
				-- if our relative position to the alarm location is the same, don't repeat the signal
				-- check if we have line of sight to the alarm point
				if (not AI.LastAlarmVec or SceneMan:ShortestDistance(AI.LastAlarmVec, AlarmVec, false).Largest > 10) then
					AI.LastAlarmVec = AlarmVec
					
					if AlarmVec.Largest < 100 then
						-- check more carfully at close range, and allow hearing of partially blocked alarm events
						if SceneMan:CastStrengthSumRay(Owner.EyePos, Event.ScenePos, 4, rte.grassID) < 100 then
							AI.AlarmPos = Vector(Event.ScenePos.X, Event.ScenePos.Y)
						end
					elseif not SceneMan:CastStrengthRay(Owner.EyePos, AlarmVec, 6, Vector(), 8, rte.grassID, true)	then
						AI.AlarmPos = Vector(Event.ScenePos.X, Event.ScenePos.Y)
					end
					
					if AI.AlarmPos then
						AI:CreateFaceAlarmBehavior(Owner)
						return true
					end
				end
			-- sometimes try to shoot back at enemies outside our view range (0.5 is the range of the brain alarm)
			elseif canSupress and Event.Range > 0.5 and PosRand() > (0.3/AI.aimSkill) and
				AlarmVec.Largest < FrameMan.PlayerScreenWidth * 1.8 and
				(not AI.LastAlarmVec or SceneMan:ShortestDistance(AI.LastAlarmVec, AlarmVec, false).Largest > 30)
			then
				-- only do this if we are facing the shortest distance to the alarm event
				local AimOwner = SceneMan:ShortestDistance(Owner.EyePos, Owner.ViewPoint, false).Normalized
				local AlarmNormal = AlarmVec.Normalized
				local dot = AlarmNormal.X * AimOwner.X + AlarmNormal.Y * AimOwner.Y
				if dot > 0.2 then
					-- check LOS
					local ID = SceneMan:CastMORay(Owner.EyePos, AlarmVec, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, false, 11)
					if ID < rte.NoMOID then
						local FoundMO = MovableMan:GetMOFromID(ID)
						if FoundMO then
							if ID ~= FoundMO.RootID then
								FoundMO = MovableMan:GetMOFromID(FoundMO.RootID)
							end
							
							if not FoundMO.EquippedItem or not FoundMO.EquippedItem:HasObjectInGroup("Explosive Weapons") then
								FoundMO = nil	-- don't shoot at without weapons or actors using tools
							end
							
							if FoundMO and FoundMO:GetController() and FoundMO:GetController():IsState(Controller.WEAPON_FIRE) and
								FoundMO.Vel.Largest < 20
							then
								-- compare the enemy aim angle with the angle of the alarm vector
								local AimEnemy = SceneMan:ShortestDistance(FoundMO.EyePos, FoundMO.ViewPoint, false).Normalized
								local dot = AlarmNormal.X * AimEnemy.X + AlarmNormal.Y * AimEnemy.Y
								if dot < -0.5 then
									-- this actor is shooting in our direction
									AI.ReloadTimer:Reset()
									AI.TargetLostTimer:Reset()
									
									-- try to shoot back
									AI.UnseenTarget = FoundMO
									AI:CreateSuppressBehavior(Owner)
									
									AI.AlarmPos = Event.ScenePos
									return true
								end
							end
						end
					else
						AI.LastAlarmVec = AlarmVec	-- don't look here again if the raycast failed
						AI.LastAlarmVec = nil
					end
				end
			end
		end
	end
end

function MeleeHumanBehaviors.GetGrenadeAngle(AimPoint, TargetVel, StartPos, muzVel)
	local Dist = SceneMan:ShortestDistance(StartPos, AimPoint, false)
	local range = Dist.Magnitude
	
	-- compensate for gravity if the point we are trying to hit is more than 2m away
	if range > 40 then
		local timeToTarget = range / muzVel
		
		-- lead the target if target speed and projectile TTT is above the threshold
		if timeToTarget * TargetVel.Magnitude > 0.5 then
			AimPoint = AimPoint + TargetVel * timeToTarget
			Dist = SceneMan:ShortestDistance(StartPos, AimPoint, false)
		end
		
		Dist = Dist / FrameMan.PPM	-- convert from pixels to meters
		local velSqr = muzVel*muzVel
		local gravity = SceneMan.GlobalAcc.Y
		local root = math.sqrt(velSqr*velSqr - gravity*(gravity*Dist.X*Dist.X+2*-Dist.Y*velSqr))
		
		if root ~= root then
			return nil	-- no solution exists if the root is NaN
		end
		
		return math.atan2(velSqr-root, gravity*Dist.X)
	end
	
	return Dist.AbsRadAngle
end

-- depricated since B30. make sure we equip our preferred device if we have one. return true if we must run this function again to be sure
function MeleeHumanBehaviors.EquipPreferredWeapon(AI, Owner)
	if AI.PlayerPreferredHD then
		Owner:EquipNamedDevice(AI.PlayerPreferredHD, true)
	elseif not Owner:EquipDeviceInGroup("Primary Weapons", true) then
		Owner:EquipDeviceInGroup("Secondary Weapons", true)
	end
	
	return false
end

-- depricated since B30. make sure we equip a primary weapon if we have one. return true if we must run this function again to be sure
function MeleeHumanBehaviors.EquipPrimaryWeapon(AI, Owner)
	Owner:EquipDeviceInGroup("Primary Weapons", true)
	return false
end

-- depricated since B30. make sure we equip a secondary weapon if we have one. return true if we must run this function again to be sure
function MeleeHumanBehaviors.EquipSecondaryWeapon(AI, Owner)
	Owner:EquipDeviceInGroup("Secondary Weapons", true)
	return false
end

-- in sentry behavior the agent only looks for new enemies, it sometimes sharp aims to increse spotting range
function MeleeHumanBehaviors.Sentry(AI, Owner)
	local sweepUp = true
	local sweepDone = false
	local maxAng = math.min(1.4, Owner.AimRange)
	local minAng = -maxAng
	local aim
	
	if AI.PlayerPreferredHD then
		Owner:EquipNamedDevice(AI.PlayerPreferredHD, true)
	elseif not Owner:EquipDeviceInGroup("Primary Weapons", true) then
		Owner:EquipDeviceInGroup("Secondary Weapons", true)
	end
	
	if AI.OldTargetPos then	-- try to reaquire an old target
		local Dist = SceneMan:ShortestDistance(Owner.EyePos, AI.OldTargetPos, false)
		AI.OldTargetPos = nil
		if (Dist.X < 0 and Owner.HFlipped) or (Dist.X > 0 and not Owner.HFlipped) then	-- we are facing the target	
			AI.deviceState = AHuman.AIMING
			AI.Ctrl.AnalogAim = Dist.Normalized
			
			for _ = 1, math.random(20, 30) do
				coroutine.yield()	-- aim here for ~0.25s
			end
		end
	elseif not AI.isPlayerOwned and Owner.AIMode ~= Actor.AIMODE_GOTO then -- face the most likely enemy approach direction
		for _ = 1, math.random(5) do	-- wait for a while
			coroutine.yield()
		end
		
		Owner:ClearMovePath()
		Owner:AddAISceneWaypoint(Vector(Owner.Pos.X, 0))
		Owner:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		-- face the direction of the first waypoint
		for WptPos in Owner.MovePath do
			local Dist = SceneMan:ShortestDistance(Owner.Pos, WptPos, false)
			if Dist.X > 5 then
				AI.SentryFacing = false
				AI.Ctrl.AnalogAim = Dist.Normalized
			elseif Dist.X < -5 then
				AI.SentryFacing = true
				AI.Ctrl.AnalogAim = Dist.Normalized
			end
			
			break
		end
		
		Owner:ClearMovePath()
	end
	
	if not AI.SentryPos then
		AI.SentryPos = Vector(Owner.Pos.X, Owner.Pos.Y)
	end
	
	while true do	-- start by looking forward
		aim = Owner:GetAimAngle(false)
		
		if sweepUp then
			if aim < maxAng/3 then
				AI.Ctrl:SetState(Controller.AIM_UP, false)
				coroutine.yield()	-- wait until next frame
				AI.Ctrl:SetState(Controller.AIM_UP, true)
			else
				sweepUp = false
			end
		else
			if aim > minAng/3 then
				AI.Ctrl:SetState(Controller.AIM_DOWN, false)
				coroutine.yield()	-- wait until next frame
				AI.Ctrl:SetState(Controller.AIM_DOWN, true)
			else
				sweepUp = true
				if sweepDone then
					break
				else
					sweepDone = true
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	if AI.SentryFacing ~= nil and Owner.HFlipped ~= AI.SentryFacing then
		Owner.HFlipped = AI.SentryFacing	-- turn to the direction we have been order to guard
		return true	-- restart this behavior
	end
	
	while true do	-- look down
		aim = Owner:GetAimAngle(false)
		if aim > minAng then
			AI.Ctrl:SetState(Controller.AIM_DOWN, true)
		else
			break
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local Hit = Vector()
	local NoObstacle = {}
	local StartPos
	AI.deviceState = AHuman.AIMING
	
	while true do	-- scan the area for obstacles
		aim = Owner:GetAimAngle(false)
		if aim < maxAng then
			AI.Ctrl:SetState(Controller.AIM_UP, true)
		else
			break
		end
		
		if Owner:EquipFirearm(false) and Owner.EquippedItem then
			StartPos = ToHeldDevice(Owner.EquippedItem).MuzzlePos
		else
			StartPos = Owner.EyePos
		end
		
		-- save the angle to a table if there is no obstacle
		if not SceneMan:CastStrengthRay(StartPos, Vector(60, 0):RadRotate(Owner:GetAimAngle(true)), 5, Hit, 2, 0, true) then
			table.insert(NoObstacle, aim)	-- TODO: don't use a table for this
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local SharpTimer = Timer()
	local aimTime = 2000
	local angDiff = 1
	AI.deviceState = AHuman.POINTING
	
	if #NoObstacle > 1 then	-- only aim where we know there are no obstacles, e.g. out of a gun port
		minAng = NoObstacle[1] * 0.95
		maxAng = NoObstacle[#NoObstacle] * 0.95
		angDiff = 1 / math.max(math.abs(maxAng - minAng), 0.1)	-- sharp aim longer from a small aiming window
	end
	
	while true do
		if not Owner:EquipFirearm(false) and not Owner:EquipThrowable(false) then
			break
		end
		
		aim = Owner:GetAimAngle(false)
		
		if sweepUp then
			if aim < maxAng then
				if aim < maxAng/5 and aim > minAng/5 and PosRand() > 0.3 then
					AI.Ctrl:SetState(Controller.AIM_UP, false)
				else
					AI.Ctrl:SetState(Controller.AIM_UP, true)
				end
			else
				sweepUp = false
			end
		else
			if aim > minAng then
				if aim < maxAng/5 and aim > minAng/5 and PosRand() > 0.3 then
					AI.Ctrl:SetState(Controller.AIM_DOWN, false)
				else
					AI.Ctrl:SetState(Controller.AIM_DOWN, true)
				end
			else
				sweepUp = true
			end
		end
		
		if SharpTimer:IsPastSimMS(aimTime) then
			SharpTimer:Reset()
			
			-- make sure that we have any preffered weapon equipped
			if AI.PlayerPreferredHD then
				Owner:EquipNamedDevice(AI.PlayerPreferredHD, true)
			elseif not Owner:EquipDeviceInGroup("Primary Weapons", true) then
				Owner:EquipDeviceInGroup("Secondary Weapons", true)
			end
			
			if AI.deviceState == AHuman.AIMING then
				aimTime = RangeRand(1000, 3000)
				AI.deviceState = AHuman.POINTING
			else
				aimTime = RangeRand(6000, 12000) * angDiff
				AI.deviceState = AHuman.AIMING
			end
			
			if Owner.AIMode ~= Actor.AIMODE_SQUAD then
				if SceneMan:ShortestDistance(Owner.Pos, AI.SentryPos, false).Magnitude > Owner.Height*0.7 then
					AI.SentryPos = SceneMan:MovePointToGround(AI.SentryPos, Owner.Height*0.25, 3)
					Owner:ClearAIWaypoints()
					Owner:AddAISceneWaypoint(AI.SentryPos)
					AI:CreateGoToBehavior(Owner)	-- try to return to the sentry pos
					break
				elseif AI.SentryFacing and Owner.HFlipped ~= AI.SentryFacing then
					Owner.HFlipped = AI.SentryFacing	-- turn to the direction we have been order to guard
					break	-- restart this behavior
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end


function MeleeHumanBehaviors.Patrol(AI, Owner)
	while AI.flying or Owner.Vel.Magnitude > 4 do	-- wait untill we are stationary
		return true
	end
	
	if Owner.ClassName == "AHuman" then
		if AI.PlayerPreferredHD then
			Owner:EquipNamedDevice(AI.PlayerPreferredHD, true)
		elseif not Owner:EquipDeviceInGroup("Primary Weapons", true) then
			Owner:EquipDeviceInGroup("Secondary Weapons", true)
		end
	end
	
	local Free = Vector()
	local WptA, WptB
	
	-- look for a path to the right
	SceneMan:CastObstacleRay(Owner.Pos, Vector(512, 0), Vector(), Free, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 4)
	local Dist = SceneMan:ShortestDistance(Owner.Pos, Free, false)
	
	if Dist.Magnitude > 20 then
		Owner:ClearAIWaypoints()
		Owner:AddAISceneWaypoint(Free)
		coroutine.yield()	-- wait until next frame
		Owner:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		local PrevPos = Vector(Owner.Pos.X, Owner.Pos.Y)
		for WptPos in Owner.MovePath do
			if math.abs(PrevPos.Y - WptPos.Y) > 14 then
				break
			end
			
			WptA = Vector(PrevPos.X, PrevPos.Y)
			PrevPos:SetXY(WptPos.X, WptPos.Y)
		end
	end
	
	-- look for a path to the left
	SceneMan:CastObstacleRay(Owner.Pos, Vector(-512, 0), Vector(), Free, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 4)
	Dist = SceneMan:ShortestDistance(Owner.Pos, Free, false)
	
	if Dist.Magnitude > 20 then
		Owner:ClearAIWaypoints()
		Owner:AddAISceneWaypoint(Free)
		coroutine.yield()	-- wait until next frame
		Owner:UpdateMovePath()
		coroutine.yield()	-- wait until next frame
		
		local PrevPos = Vector(Owner.Pos.X, Owner.Pos.Y)
		for WptPos in Owner.MovePath do
			if math.abs(PrevPos.Y - WptPos.Y) > 14 then
				break
			end
			
			WptB = Vector(PrevPos.X, PrevPos.Y)
			PrevPos:SetXY(WptPos.X, WptPos.Y)
		end
	end	
	
	Owner:ClearAIWaypoints()
	coroutine.yield()	-- wait until next frame
	
	if WptA then
		Dist = SceneMan:ShortestDistance(Owner.Pos, WptA, false)
		if Dist.Magnitude > 20 then
			Owner:AddAISceneWaypoint(WptA)
		else
			WptA = nil
		end
	end
	
	if WptB then
		Dist = SceneMan:ShortestDistance(Owner.Pos, WptB, false)
		if Dist.Magnitude > 20 then
			Owner:AddAISceneWaypoint(WptB)
		else
			WptB = nil
		end
	end
	
	if WptA or WptB then
		AI:CreateGoToBehavior(Owner)
	else	-- no path was found
		local FlipTimer = Timer()
		FlipTimer:SetSimTimeLimitMS(3000)
		while true do
			coroutine.yield()	-- wait until next frame
			if FlipTimer:IsPastSimTimeLimit() then
				FlipTimer:Reset()
				FlipTimer:SetSimTimeLimitMS(RangeRand(2000, 5000))
				Owner.HFlipped = not Owner.HFlipped	-- turn around and try the other direction sometimes
				if PosRand() < 0.3 then
					break	-- end the behavior
				end
			end
		end
	end
	
	return true
end


function MeleeHumanBehaviors.GoldDig(AI, Owner)
	-- make sure our weapon have ammo before we start to dig, just in case we encounter an enemy while digging
	if Owner.EquippedItem and (Owner.FirearmNeedsReload or Owner.FirearmIsEmpty) and Owner.EquippedItem:HasObjectInGroup("Weapons") then
		Owner:ReloadFirearm()
		
		repeat
			coroutine.yield()	-- wait until next frame
		until not Owner.FirearmIsEmpty
	end
	
	-- select a digger
	if not Owner:EquipDiggingTool(true) then
		return true -- our digger is gone, abort this behavior
	end
	
	local aimAngle = 0.45
	local BestGoldLocation = {X = 0, Y = 0}
	local smallestPenalty = math.huge
	
	for aimAngle = 0.4, -3.54, -0.033 do
		local Digger
		if Owner.EquippedItem then
			Digger = ToHeldDevice(Owner.EquippedItem)
			if not Digger then
				break
			end
		else
			break
		end
		
		local LookVec
		if aimAngle < -0.8 and aimAngle > -2.4 then
			LookVec = Vector(60,0):RadRotate(aimAngle)
		else	-- search further away horizontaly
			LookVec = Vector(180,0):RadRotate(aimAngle)
		end
		
		AI.Ctrl.AnalogAim = LookVec.Normalized
		local GoldPos = Vector()
		if SceneMan:CastMaterialRay(Digger.MuzzlePos, LookVec, rte.goldID, GoldPos, 1, true) then
			-- avoid gold close to the edeges of the scene
			if GoldPos.Y < SceneMan.SceneHeight - 25 and
				(SceneMan.SceneWrapsX or (GoldPos.X > 50 and GoldPos.X < SceneMan.SceneWidth - 50))
			then
				local Dist = SceneMan:ShortestDistance(Owner.Pos, GoldPos, false)	-- prioritize gold close to us
				local str = SceneMan:CastStrengthSumRay(Owner.EyePos, GoldPos, 3, rte.goldID) / 30	-- prioritize gold in soft ground
				local penalty = str + Dist.Magnitude + math.abs(Dist.Y*5)
				local DigArea = SceneMan:ShortestDistance(GoldPos, Owner.EyePos+LookVec, false)
				local digLength = math.min(DigArea.Magnitude, 180)  -- sanity check to circumvent infinite loops
				
				-- prioritize gold located horizontaly or below us
				if Dist.Y > -20 then
					penalty = penalty - 5
				end
				
				coroutine.yield()	-- wait until next frame
				
				-- prioritize areas with more gold
				DigArea:Normalize()
				for i = 5, digLength, 5 do
					local Step = GoldPos + DigArea * i
					if Step.X >= SceneMan.SceneWidth then
						if SceneMan.SceneWrapsX then
							Step.X = Step.X - SceneMan.SceneWidth
						else
							break
						end
					elseif Step.X < 0 then
						if SceneMan.SceneWrapsX then
							Step.X = SceneMan.SceneWidth - Step.X
						else
							break
						end
					end
					
					if Step.Y > SceneMan.SceneHeight - 50 then
						break
					end
					
					if SceneMan:GetTerrMatter(Step.X, Step.Y) == rte.goldID then
						penalty = penalty - 4
					end
				end
				
				-- prioritize gold located horizontaly relative to us
				if math.abs(Dist.X) > math.abs(Dist.Y) then
					if math.abs(Dist.X) * 0.5 > math.abs(Dist.Y) then
						penalty = penalty - 80
					else
						penalty = penalty - 40
					end
				end
				
				if penalty < smallestPenalty then
					if Dist.Magnitude < 50 then	-- dig to a point behind the gold
						GoldPos = Owner.Pos + Dist:SetMagnitude(55)
					end
					
					-- make sure there is no metal in our path
					if not SceneMan:CastStrengthRay(Owner.Pos, Dist:SetMagnitude(60), 95, Vector(), 2, rte.grassID, SceneMan.SceneWrapsX) then
						smallestPenalty = penalty + RangeRand(-7, 7)
						BestGoldLocation.X, BestGoldLocation.Y = GoldPos.X, GoldPos.Y
					end
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	local BestGoldPos = Vector(BestGoldLocation.X, BestGoldLocation.Y)
	if BestGoldPos.Largest == 0 then
		if Owner.Pos.Y < SceneMan.SceneHeight - 50 then	-- don't dig beyond the scene limit
			-- no gold found, so dig down and try again
			local rayLenghtY = math.min(80, SceneMan.SceneHeight-100)
			local rayLenghtX = rayLenghtY * 0.5
			local Target = Owner.Pos + Vector(rayLenghtX, rayLenghtY)
			local str_r = SceneMan:CastStrengthSumRay(Owner.Pos, Target, 6, rte.goldID)
			coroutine.yield()	-- wait until next frame
			
			Target = Owner.Pos + Vector(-rayLenghtX, rayLenghtY)
			local str_l = SceneMan:CastStrengthSumRay(Owner.Pos, Target, 6, rte.goldID)
			coroutine.yield()	-- wait until next frame
			
			if str_r < str_l then
				BestGoldPos = Owner.Pos + Vector(rayLenghtX, rayLenghtY)
			else
				BestGoldPos = Owner.Pos + Vector(-rayLenghtX, rayLenghtY)
			end
		else
			-- no gold here, and we cannot dig deeper, calculate average horizontal strength
			local rayLenght = 80
			local Target = Owner.Pos + Vector(rayLenght, -5)
			local Trace = SceneMan:ShortestDistance(Owner.Pos, Target, false)
			local str_r = SceneMan:CastStrengthSumRay(Owner.Pos, Target, 5, rte.goldID)
			local obst_r = SceneMan:CastStrengthRay(Owner.Pos, Trace, 95, Vector(), 2, rte.grassID, SceneMan.SceneWrapsX)
			coroutine.yield()	-- wait until next frame
			
			Target = Owner.Pos + Vector(-rayLenght, -5)
			Trace = SceneMan:ShortestDistance(Owner.Pos, Target, false)
			local str_l = SceneMan:CastStrengthSumRay(Owner.Pos, Target, 5, rte.goldID)
			local obst_l = SceneMan:CastStrengthRay(Owner.Pos, Trace, 95, Vector(), 2, rte.grassID, SceneMan.SceneWrapsX)
			coroutine.yield()	-- wait until next frame
			
			local goLeft
			if obst_l then
				goLeft = false
			elseif obst_r then
				goLeft = true
			else
				goLeft = math.random() > 0.5
				
				-- go towards the larger obstacle, unless metal
				if math.abs(str_l - str_r) > 200 then
					if str_r > str_l and not obst_r then
						goLeft = false
					elseif str_r < str_l and not obst_l then
						goLeft = true
					end
				end
			end
		
			if goLeft then
				BestGoldPos = Owner.Pos + Vector(-rayLenght, -5)
			else
				BestGoldPos = Owner.Pos + Vector(rayLenght, -5)
			end
		end
	end
	
	BestGoldPos.Y = math.min(BestGoldPos.Y, SceneMan.SceneHeight-30)
	Owner:ClearAIWaypoints()
	Owner:AddAISceneWaypoint(BestGoldPos)
	AI:CreateGoToBehavior(Owner)
	
	return true
end


-- find the closest enemy brain
function MeleeHumanBehaviors.BrainSearch(AI, Owner)
	if AI.PlayerPreferredHD then
		Owner:EquipNamedDevice(AI.PlayerPreferredHD, true)
	end
	
	local Brains = {}
	for Act in MovableMan.Actors do
		if Act.Team ~= Owner.Team and Act:HasObjectInGroup("Brains") then
			table.insert(Brains, Act)
		end
	end
	
	if #Brains < 1 then	-- no bain actors found, check if some other actor is the brain
		local GmActiv = ActivityMan:GetActivity()
		for player = Activity.PLAYER_1, Activity.MAXPLAYERCOUNT - 1 do
			if GmActiv:PlayerActive(player) and GmActiv:GetTeamOfPlayer(player) ~= Owner.Team then
				local Act = GmActiv:GetPlayerBrain(player)
				if Act and MovableMan:IsActor(Act) then
					table.insert(Brains, Act)
				end
			end
		end
	end
	
	if #Brains > 0 then
		coroutine.yield()	-- wait until next frame
		
		if #Brains == 1 then
			if MovableMan:IsActor(Brains[1]) then
				Owner:ClearAIWaypoints()
				Owner:AddAIMOWaypoint(Brains[1])
				AI:CreateGoToBehavior(Owner)
			end
		else
			local ClosestBrain
			local minDist = math.huge
			for _, Act in pairs(Brains) do
				-- measure how easy the path to the destination is to traverse
				if MovableMan:IsActor(Act) then
					Owner:ClearAIWaypoints()
					Owner:AddAISceneWaypoint(Act.Pos)
					Owner:UpdateMovePath()
					
					local OldWpt, deltaY
					local index = 0
					local height = 0
					local pathLength = 0
					local pathObstMaxHeight = 0
					for Wpt in Owner.MovePath do
						pathLength = pathLength + 1
						if OldWpt then
							deltaY = OldWpt.Y - Wpt.Y
							if deltaY > 20 then	-- Wpt is more than n pixels above OldWpt in the scene
								if deltaY / math.abs(SceneMan:ShortestDistance(OldWpt, Wpt, false).X) > 1 then	-- the slope is more than 45 degrees
									height = height + (OldWpt.Y - Wpt.Y)
									pathObstMaxHeight = math.max(pathObstMaxHeight, height)
								else
									height = 0
								end
							else
								height = 0
							end
						end
						
						OldWpt = Wpt
						
						if index > 20 then
							index = 0
							coroutine.yield()	-- wait until the next frame
						else
							index = index + 1
						end
					end
					
					local score = pathLength * 0.55 + math.floor(pathObstMaxHeight/27) * 8
					if score < minDist then
						minDist = score
						ClosestBrain = Act
					end
					
					coroutine.yield()	-- wait until next frame
				end
			end
			
			Owner:ClearAIWaypoints()
			
			if MovableMan:IsActor(ClosestBrain) then
				Owner:AddAIMOWaypoint(ClosestBrain)
				AI:CreateGoToBehavior(Owner)
			else
				return true	-- the brain we found died while we where searching, restart this behavior next frame
			end
		end
	else	-- no enemy brains left
		AI:CreateSentryBehavior(Owner)
	end
	
	return true
end


-- find a weapon to pick up
function MeleeHumanBehaviors.WeaponSearch(AI, Owner)
	local minDist
	local Devices = {}
	local pickupDiggers = not Owner:HasObjectInGroup("Diggers")
	
	if AI.isPlayerOwned then
		minDist = 100	-- don't move player actors more than 4m
	else
		minDist = FrameMan.PlayerScreenWidth * 0.45
	end
	
	if Owner.AIMode == Actor.AIMODE_SENTRY then
		minDist = minDist * 0.6
	end
	
	local itemsFound = 0
	for Item in MovableMan.Items do	-- store all HeldDevices of the correct type and within a cerain range in a table
		local HD = ToHeldDevice(Item)
		if HD and not HD:IsActivated() and HD.Vel.Largest < 3 and
			SceneMan:ShortestDistance(Owner.Pos, HD.Pos, false).Largest < minDist and
			not SceneMan:IsUnseen(HD.Pos.X, HD.Pos.Y, Owner.Team)
		then
			table.insert(Devices, HD)
			itemsFound = itemsFound + 1
		end
	end
	
	if itemsFound > 0 then
		coroutine.yield()	-- wait until next frame
		
		if AI.isPlayerOwned then
			minDist = 10	-- # of waypoints
		else
			minDist = 36
		end
		
		local waypoints, score
		local DevicesToPickUp = {}
		for _, Item in pairs(Devices) do
			if MovableMan:ValidMO(Item) then
				waypoints = SceneMan.Scene:CalculatePath(Owner.Pos, Item.Pos, false, 1)
				if waypoints < minDist and waypoints > -1 then
					-- estimate the walking distance to the item
					if Item:HasObjectInGroup("Primary Weapons") then
						score = waypoints * 0.4	-- prioritize primaries
					elseif Item.ClassName == "TDExplosive" then
						score = waypoints * 1.4	-- avoid grenades if there are other weapons
					elseif Item:IsTool() then
						if pickupDiggers and Item:HasObjectInGroup("Diggers") then
							score = waypoints * 1.8	-- avoid diggers if there are other weapons
						else
							waypoints = minDist -- don't pick up
						end
					else
						score = waypoints
					end
					
					if waypoints < minDist then
						table.insert(DevicesToPickUp, {HD=Item, score=score})
					end
				end
				
				coroutine.yield()	-- wait until next frame
			end
		end
		
		AI.PickupHD = nil
		table.sort(DevicesToPickUp, function(A,B) return A.score < B.score end)	-- sort the items in order of discounted distance
		for _, Data in pairs(DevicesToPickUp) do
			if MovableMan:ValidMO(Data.HD) and Data.HD:IsDevice() then
				AI.PickupHD = Data.HD
				break
			end
		end
		
		if AI.PickupHD then
			-- where do we move after pick up?
			local PrevMoveTarget, PrevSeceneWaypoint
			if Owner.MOMoveTarget and MovableMan:ValidMO(Owner.MOMoveTarget) then
				PrevMoveTarget = Owner.MOMoveTarget
			else
				PrevSeceneWaypoint = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height/5, 4)	-- last wpt or current pos
			end
			
			Owner:ClearMovePath()
			Owner:AddAIMOWaypoint(AI.PickupHD)
			
			if PrevMoveTarget then
				Owner:AddAIMOWaypoint(PrevMoveTarget)
			elseif PrevSeceneWaypoint then
				Owner:AddAISceneWaypoint(PrevSeceneWaypoint)
			end
			
			if Owner.AIMode == Actor.AIMODE_SENTRY then
				AI.SentryFacing = Owner.HFlipped
			end
			
			Owner:UpdateMovePath()
			AI:CreateGoToBehavior(Owner)
		end
	end
	
	return true
end


-- find a tool to pick up
function MeleeHumanBehaviors.ToolSearch(AI, Owner)
	local minDist
	if Owner.AIMode == Actor.AIMODE_GOLDDIG then
		minDist = FrameMan.PlayerScreenWidth * 0.5	-- move up to half a screen when digging
	elseif AI.isPlayerOwned then
		minDist = 60	-- don't move player actors more than 3m
	else
		minDist = FrameMan.PlayerScreenWidth * 0.3
	end
	
	if Owner.AIMode == Actor.AIMODE_SENTRY then
		minDist = minDist * 0.6
	end
	
	local Devices = {}
	local itemsFound = 0
	for Item in MovableMan.Items do	-- store all HeldDevices of the correct type and within a cerain range in a table
		local HD = ToHeldDevice(Item)
		if HD and not HD:IsActivated() and HD.Vel.Largest < 3 and
			SceneMan:ShortestDistance(Owner.Pos, HD.Pos, false).Largest < minDist and
			not SceneMan:IsUnseen(HD.Pos.X, HD.Pos.Y, Owner.Team)
		then
			table.insert(Devices, HD)
			itemsFound = itemsFound + 1
		end
	end
	
	if itemsFound > 0 then
		coroutine.yield()	-- wait until next frame
		
		if Owner.AIMode == Actor.AIMODE_GOLDDIG then
			minDist = 30
		elseif AI.isPlayerOwned then
			minDist = 5
		else
			minDist = 16
		end
		
		local DevicesToPickUp = {}
		for _, Item in pairs(Devices) do
			if MovableMan:ValidMO(Item) and Item:HasObjectInGroup("Diggers") then
				-- estimate the walking distance to the item
				local waypoints = SceneMan.Scene:CalculatePath(Owner.Pos, Item.Pos, false, 1)
				if waypoints < minDist and waypoints > -1 then
					table.insert(DevicesToPickUp, {HD=Item, score=waypoints})
				end
				
				coroutine.yield()	-- wait until next frame
			end
		end
		
		AI.PickupHD = nil
		table.sort(DevicesToPickUp, function(A,B) return A.score < B.score end)	-- sort the items in order of waypoints
		for _, Data in pairs(DevicesToPickUp) do
			if MovableMan:ValidMO(Data.HD) and Data.HD:IsDevice() then
				AI.PickupHD = Data.HD
				break
			end
		end
		
		if AI.PickupHD then
			-- where do we move after pick up?
			local PrevMoveTarget, PrevSeceneWaypoint
			if Owner.MOMoveTarget and MovableMan:ValidMO(Owner.MOMoveTarget) then
				PrevMoveTarget = Owner.MOMoveTarget
			else
				PrevSeceneWaypoint = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height/5, 4)	-- last wpt or current pos
			end
			
			Owner:ClearMovePath()
			Owner:AddAIMOWaypoint(AI.PickupHD)
			
			if Owner.AIMode ~= Actor.AIMODE_GOLDDIG then
				if PrevMoveTarget then
					Owner:AddAIMOWaypoint(PrevMoveTarget)
				elseif PrevSeceneWaypoint then
					Owner:AddAISceneWaypoint(PrevSeceneWaypoint)
				end
				
				if Owner.AIMode == Actor.AIMODE_SENTRY then
					AI.SentryFacing = Owner.HFlipped
				end
			end
			
			Owner:UpdateMovePath()
			AI:CreateGoToBehavior(Owner)
		end
	end
	
	return true
end


-- move to the next waypoint
function MeleeHumanBehaviors.GoToWpt(AI, Owner)
	-- check if we have arrived
	if not (Owner.AIMode == Actor.AIMODE_SQUAD or Owner:GetWaypointListSize() > 0) then
		if not Owner.MOMoveTarget then
			if SceneMan:ShortestDistance(Owner:GetLastAIWaypoint(), Owner.Pos, false).Largest < Owner.Height * 0.15 then			
				Owner:ClearAIWaypoints()
				Owner:ClearMovePath()
				Owner:DrawWaypoints(false)
				AI:CreateSentryBehavior(Owner)
				
				if Owner.AIMode == Actor.AIMODE_GOTO then
					AI.SentryFacing = Owner.HFlipped	-- guard this direction
					AI.SentryPos = Vector(Owner.Pos.X, Owner.Pos.Y) -- guard this point
				end
				
				return true
			end
		end
	end
	
	-- is Y1 lower down in the scene, compared to Y2?
	local Lower = function(Y1, Y2, margin)
		return Y1.Pos.Y - margin > Y2.Pos.Y
	end
	
	local ArrivedTimer = Timer()
	local UpdatePathTimer = Timer()
	
	if Owner.MOMoveTarget then
		UpdatePathTimer:SetSimTimeLimitMS(RangeRand(7000, 8000))
	else
		UpdatePathTimer:SetSimTimeLimitMS(RangeRand(12000, 14000))
	end
	
	local NoLOSTimer = Timer()
	NoLOSTimer:SetSimTimeLimitMS(1000)
	
	local StuckTimer = Timer()
	StuckTimer:SetSimTimeLimitMS(3000)
	
	local nextLatMove = AI.lateralMoveState
	local nextAimAngle = Owner:GetAimAngle(false) * 0.95
	local scanAng = 0	-- for obstacle detection
	local Obstacles = {}
	local PrevWptPos = Vector(Owner.Pos.X, Owner.Pos.Y)
	local sweepCW = true
	local sweepRange = 0
	local digState = AHuman.NOTDIGGING
	local obstacleState = Actor.PROCEEDING
	local WptList, Waypoint, Dist, CurrDist
	local Obst = {R_LOW = 1, R_FRONT = 2, R_HIGH = 3, R_UP = 5, L_UP = 6, L_HIGH = 8, L_FRONT = 9, L_LOW = 10}
	local Facings = {{aim=0, facing=0}, {aim=1.4, facing=1.4}, {aim=1.4, facing=math.pi-1.4}, {aim=0, facing=math.pi}}
	
	while true do
		if Owner.Vel.Largest > 2 then
			StuckTimer:Reset()
		end
		
		if AI.refuel and Owner.Jetpack then
			-- if jetpack is full or we are falling we can stop refueling
			if Owner.JetTimeLeft > Owner.JetTimeTotal * 0.98 or
				(AI.flying and Owner.Vel.Y < -3 and Owner.JetTimeLeft > AI.minBurstTime*2)
			then
				AI.refuel = false
			elseif not AI.flying then
				AI.jump = false
				AI.lateralMoveState = Actor.LAT_STILL
			end
		elseif not AI.flying and UpdatePathTimer:IsPastSimTimeLimit() then
			UpdatePathTimer:Reset()
			
			if Waypoint and AI.BlockingMO then
				if MovableMan:ValidMO(AI.BlockingMO) then
					CurrDist = SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false)
					if (Owner.Pos.X > AI.BlockingMO.Pos.X and CurrDist.X < Owner.Pos.X) or
						(Owner.Pos.X < AI.BlockingMO.Pos.X and CurrDist.X > Owner.Pos.X) or
						SceneMan:ShortestDistance(Owner.Pos, AI.BlockingMO.Pos, false).Magnitude > Owner.Diameter + AI.BlockingMO.Diameter
					then
						AI.BlockingMO = nil	-- the blocking actor is not in the way any longer
						AI.teamBlockState = Actor.NOTBLOCKED
					else
						AI.BlockedTimer:Reset()
						AI.teamBlockState = Actor.IGNORINGBLOCK
						AI:CreateMoveAroundBehavior(Owner)
						break	-- end this behavior
					end
				else
					AI.BlockingMO = nil
				end
			end
			
			AI.deviceState = AHuman.STILL
			AI.proneState = AHuman.NOTPRONE
			AI.jump = false
			nextLatMove = Actor.LAT_STILL
			digState = AHuman.NOTDIGGING
			Waypoint = nil
			WptList = nil -- update the path
		elseif StuckTimer:IsPastSimTimeLimit() then	-- dislodge
			if AI.jump then
				if Owner.Jetpack and Owner.JetTimeLeft < AI.minBurstTime then	-- out of fuel
					AI.jump = false
					AI.refuel = true
					nextLatMove = Actor.LAT_STILL
				else
					local chance = PosRand()
					if chance < 0.1 then
						nextLatMove = Actor.LAT_LEFT
					elseif chance > 0.9 then
						nextLatMove = Actor.LAT_RIGHT
					else
						nextLatMove = Actor.LAT_STILL
					end
				end
			else
				if PosRand() < 0.2 then
					if AI.lateralMoveState == Actor.LAT_LEFT then
						nextLatMove = Actor.LAT_RIGHT
					else
						nextLatMove = Actor.LAT_LEFT
					end
				end
				
				-- refueling done
				if AI.refuel and Owner.Jetpack and Owner.JetTimeLeft >= Owner.JetTimeTotal * 0.99 then
					AI.jump = true
				end
			end
		elseif WptList then	-- we have a list of waypoints, follow it
			if not WptList[1] and not Waypoint then	-- arrived
				if Owner.MOMoveTarget then -- following actor
					if Owner.MOMoveTarget:IsActor() then
						local Trace = SceneMan:ShortestDistance(Owner.Pos, Owner.MOMoveTarget.Pos, false)
						if Trace.Largest < Owner.Height * 0.5 + (Owner.MOMoveTarget.Height or 100) * 0.5 and
							SceneMan:CastStrengthRay(Owner.Pos, Trace, 5, Vector(), 4, rte.grassID, true)
						then -- add a waypoint if the MOMoveTarget is close and in LOS
							Waypoint = {Pos=SceneMan:MovePointToGround(Owner.MOMoveTarget.Pos, Owner.Height*0.2, 4)}
						else
							WptList = nil -- update the path
						end
					end
				else	-- moving towards a scene point
					--local GroundPos = Owner:GetLastAIWaypoint()
					local GroundPos = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height*0.2, 4)
					if SceneMan:ShortestDistance(GroundPos, Owner.Pos, false).Largest < Owner.Height * 0.4 then
						if Owner.AIMode == Actor.AIMODE_GOTO then
							AI.SentryFacing = Owner.HFlipped	-- guard this direction
							AI.SentryPos = Vector(Owner.Pos.X, Owner.Pos.Y) -- guard this point
							AI:CreateSentryBehavior(Owner)
						end
						
						Owner:ClearAIWaypoints()
						Owner:ClearMovePath()
						Owner:DrawWaypoints(false)
						return true
					end
				end
			else
				if not Waypoint then	-- get the next waypoint in the list
					if WptList then
						local NextWptPos = WptList[1].Pos
						Dist = SceneMan:ShortestDistance(Owner.Pos, NextWptPos, false)
						if Dist.Y < -25 and math.abs(Dist.X) < 30 then	-- avoid any corners if the next waypoint is above us
							local cornerType
							local CornerPos = Vector(NextWptPos.X, NextWptPos.Y)
							if Owner.Pos.X > CornerPos.X then
								CornerPos = CornerPos + Vector(25, -40)
								cornerType = "right"
							else
								CornerPos = CornerPos + Vector(-25, -40)
								cornerType = "left"
							end
							
							local Free = Vector()
							Dist = SceneMan:ShortestDistance(NextWptPos, CornerPos, false)
							-- make sure the corner waypoint is not inside terrain
							local pixels = SceneMan:CastObstacleRay(NextWptPos, Dist, Vector(), Free, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 3)
							if pixels == 0 then
								break	-- the waypoint is inside terrain, plot a new path
							elseif pixels > 0 then
								CornerPos = (NextWptPos + Free) / 2	-- compensate for obstacles
							end
							
							coroutine.yield()	-- wait until next frame
							
							-- check if we have LOS
							Dist = SceneMan:ShortestDistance(Owner.Pos, CornerPos, false)
							if 0 <= SceneMan:CastObstacleRay(Owner.Pos, Dist, Vector(), Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 2) then
								-- CornerPos is blocked
								CornerPos.X = Owner.Pos.X	-- move CornerPos straight above us
								cornerType = "air"
							end
							
							coroutine.yield()	-- wait until next frame
							
							Waypoint = {Pos=CornerPos, Type=cornerType}
							if WptList[2] and not WptList[1].Type then	-- remove the waypoint after the corner if possible
								table.remove(WptList, 1)
								Owner:RemoveMovePathBeginning() -- clean up the graphical representation of the path
							end
							
							if not Owner.MOMoveTarget then
								Owner:AddToMovePathBeginning(Waypoint.Pos)
							end
						else
							Waypoint = table.remove(WptList, 1)
							if Waypoint.Type ~= "air" then
								local Free = Vector()
								
								-- only if we have a digging tool
								if Waypoint.Type ~= "drop" and Owner:HasObjectInGroup("Diggers") then
									local PathSegRay = SceneMan:ShortestDistance(PrevWptPos, Waypoint.Pos, false)	-- detect material blocking the path and start digging through it
									if AI.teamBlockState ~= Actor.BLOCKED and SceneMan:CastStrengthRay(PrevWptPos, PathSegRay, 4, Free, 2, rte.doorID, true) then
										if SceneMan:ShortestDistance(Owner.Pos, Free, false).Magnitude < Owner.Height*0.4 then	-- check that we're close enough to start digging
											digState = AHuman.STARTDIG
											AI.deviceState = AHuman.DIGGING
											obstacleState = Actor.DIGPAUSING
											nextLatMove = Actor.LAT_STILL
											sweepRange = math.min(math.pi*0.2, Owner.AimRange)
											StuckTimer:SetSimTimeLimitMS(6000)
											AI.Ctrl.AnalogAim = SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false).Normalized	-- aim in the direction of the next waypoint
										else
											digState = AHuman.NOTDIGGING
											obstacleState = Actor.PROCEEDING
										end
										
										coroutine.yield()	-- wait until next frame
									else
										digState = AHuman.NOTDIGGING
										obstacleState = Actor.PROCEEDING
										StuckTimer:SetSimTimeLimitMS(2000)
									end
								end
								
								if digState == AHuman.NOTDIGGING and AI.deviceState ~= AHuman.DIGGING then
									-- if our path isn't blocked enough to dig, but the headroom is too little, start crawling to get through
									local Heading = SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false):SetMagnitude(Owner.Height*0.5)
									
									-- don't crawl if it's too steep, climb then instead
									if math.abs(Heading.X) > math.abs(Heading.Y) and Owner.Head and Owner.Head:IsAttached() then
										local TopHeadPos = Owner.Head.Pos - Vector(0, Owner.Head.Radius*0.7)
										
										-- first check up to the top of the head, and then from there forward
										if SceneMan:CastStrengthRay(Owner.Pos, TopHeadPos - Owner.Pos, 5, Free, 4, rte.doorID, true) or
												SceneMan:CastStrengthRay(TopHeadPos, Heading, 5, Free, 4, rte.doorID, true)
										then
											AI.proneState = AHuman.PRONE
										else
											AI.proneState = AHuman.NOTPRONE
										end
										
										coroutine.yield()	-- wait until next frame
									else
										AI.proneState = AHuman.NOTPRONE
									end
								end
							end
							
							if not WptList[1] then
								WptList = nil -- update the path
							end
						end
						
						if not Waypoint.Type then
							ArrivedTimer:SetSimTimeLimitMS(100)
						elseif Waypoint.Type == "last" then
							ArrivedTimer:SetSimTimeLimitMS(600)
						else	-- air or corner wpt
							ArrivedTimer:SetSimTimeLimitMS(25)
						end
					end
				elseif WptList[2] then	-- check if some other waypoint is closer
					local test = math.random(1, math.min(10, #WptList))
					local RandomWpt = WptList[test]
					if RandomWpt then
						Dist = SceneMan:ShortestDistance(Owner.Pos, RandomWpt.Pos, false)
						local mag = Dist.Magnitude
						if mag < 50 and mag < SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false).Magnitude/3 then
							-- this waypoint is closer, check LOS
							if -1 == SceneMan:CastObstacleRay(Owner.Pos, Dist, Vector(), Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 4) then
								Waypoint = RandomWpt	-- go here instead
								if WptList[test-1] then
									PrevWptPos = Vector(WptList[test-1].Pos.X, WptList[test-1].Pos.Y)
								else
									PrevWptPos = Vector(Owner.Pos.X, Owner.Pos.Y)
								end
								
								for _ = 1, test do	-- delete the earlier waypoints
									table.remove(WptList, 1)
									if WptList[1] then
										Owner:RemoveMovePathBeginning()
									end
								end
							end
							
							if not AI.jump and not AI.flying then
								coroutine.yield()	-- wait until next frame
							end
						end
					end
				end
				
				if Waypoint then
					if not WptList and Owner.MOMoveTarget and MovableMan:ValidMO(Owner.MOMoveTarget) then
						local Trace = SceneMan:ShortestDistance(Owner.Pos, Owner.MOMoveTarget.Pos, false)
						
						if Owner.MOMoveTarget.Team == Owner.Team then
							if Trace.Largest > Owner.Height * 0.3 + (Owner.MOMoveTarget.Height or 100) * 0.3 then
								Waypoint.Pos = Owner.MOMoveTarget.Pos
							else	-- arrived
								if not AI.flying then
									while true do
										StuckTimer:Reset()
										UpdatePathTimer:Reset()
										AI.lateralMoveState = Actor.LAT_STILL
										AI.jump = false
										coroutine.yield()	-- wait until next frame
										
										if Owner.MOMoveTarget and MovableMan:ValidMO(Owner.MOMoveTarget) then
											Trace = SceneMan:ShortestDistance(Owner.Pos, Owner.MOMoveTarget.Pos, false)
											if Trace.Largest > Owner.Height * 0.4 + (Owner.MOMoveTarget.Height or 100) * 0.4 or
												SceneMan:CastStrengthRay(Owner.Pos, Trace, 5, Vector(), 4, rte.doorID, true)
											then
												Waypoint = nil
												WptList = nil -- update the path
												break
											end
											
											coroutine.yield()	-- wait until next frame
										else -- MOMoveTarget gone
											return true
										end
									end
								end
							end
						elseif Trace.Largest < Owner.Height * 0.33 + (Owner.MOMoveTarget.Height or 100) * 0.33 then -- enemy MO
							Waypoint.Pos = Owner.MOMoveTarget.Pos
						end
					end
					
					if Waypoint then
						CurrDist = SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false)
						
						-- digging
						if digState ~= AHuman.NOTDIGGING then
							if not AI.Target and Owner:EquipDiggingTool(true) then	-- switch to the digger if we have one
								if Owner.FirearmIsEmpty then	-- reload if it's empty
									AI.fire = false
									AI.Ctrl:SetState(Controller.WEAPON_RELOAD, true)
								else
									if AI.teamBlockState == Actor.BLOCKED then
										AI.fire = false
										nextLatMove = Actor.LAT_STILL
									else
										if obstacleState == Actor.PROCEEDING then
											if CurrDist.X < -1 then
												nextLatMove = Actor.LAT_LEFT
											elseif CurrDist.X > 1 then
												nextLatMove = Actor.LAT_RIGHT
											end
										else
											nextLatMove = Actor.LAT_STILL
										end
										
										-- check if we are close enough to dig
										if SceneMan:ShortestDistance(PrevWptPos, Owner.Pos, false).Magnitude > Owner.Height*0.5 and
											 SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false).Magnitude > Owner.Height*0.5
										then
											digState = AHuman.NOTDIGGING
											obstacleState = Actor.PROCEEDING
											AI.deviceState = AHuman.STILL
											AI.fire = false
											Owner:EquipFirearm(true)
										else
											-- see if we have dug out all that we can in the sweep area without moving closer
											local centerAngle = CurrDist.AbsRadAngle
											local Ray = Vector(Owner.Height*0.3, 0):RadRotate(centerAngle)	-- center
											if SceneMan:CastNotMaterialRay(Owner.Pos, Ray, 0, 3, false) < 0 then
												coroutine.yield()	-- wait until next frame
												
												-- now check the tunnel's thickness
												Ray = Vector(Owner.Height*0.3, 0):RadRotate(centerAngle + sweepRange)	-- up
												if SceneMan:CastNotMaterialRay(Owner.Pos, Ray, rte.airID, 3, false) < 0 then
													coroutine.yield()	-- wait until next frame
													
													Ray = Vector(Owner.Height*0.3, 0):RadRotate(centerAngle - sweepRange)	-- down
													if SceneMan:CastNotMaterialRay(Owner.Pos, Ray, rte.airID, 3, false) < 0 then
														obstacleState = Actor.PROCEEDING	-- ok the tunnel section is clear, so start walking forward while still digging
													else
														obstacleState = Actor.DIGPAUSING	-- tunnel cavity not clear yet, so stay put and dig some more
													end
												end
											else
												obstacleState = Actor.DIGPAUSING	-- tunnel cavity not clear yet, so stay put and dig some more
											end
											
											coroutine.yield()	-- wait until next frame
											
											local aimAngle = Owner:GetAimAngle(true)
											local AimVec = Vector(1, 0):RadRotate(aimAngle)
											
											local angDiff = math.asin(AimVec:Cross(CurrDist.Normalized))	-- the angle between CurrDist and AimVec
											if math.abs(angDiff) < sweepRange then
												AI.fire = true	-- only fire the digger at the obstacle
											else
												AI.fire = false
											end
											
											-- sweep the digger between the two endpoints of the obstacle
											local DigTarget
											if sweepCW then
												DigTarget = Vector(Owner.Height*0.4, 0):RadRotate(centerAngle + sweepRange)
											else
												DigTarget = Vector(Owner.Height*0.4, 0):RadRotate(centerAngle - sweepRange)
											end
											
											angDiff = math.asin(AimVec:Cross(DigTarget.Normalized))	-- The angle between DigTarget and AimVec
											if math.abs(angDiff) < 0.1 then
												sweepCW = not sweepCW	-- this is close enough, go in the other direction next frame
											else
												AI.Ctrl.AnalogAim = (Vector(AimVec.X, AimVec.Y):RadRotate(-angDiff*0.15)).Normalized
											end
											
											-- check if we are done when we get close enough to the waypoint
											if SceneMan:ShortestDistance(Owner.Pos, Waypoint.Pos, false).Largest < Owner.Height*0.3 then
												if not SceneMan:CastStrengthRay(PrevWptPos, SceneMan:ShortestDistance(PrevWptPos, Waypoint.Pos, false), 5, Vector(), 1, rte.doorID, true) and
													not SceneMan:CastStrengthRay(Owner.EyePos, SceneMan:ShortestDistance(Owner.EyePos, Waypoint.Pos, false), 5, Vector(), 1, rte.doorID, true)
												then
													-- advance to the next waypoint, if there are any
													if WptList and WptList[1] then
														UpdatePathTimer:Reset()
														PrevWptPos = Waypoint.Pos
														Waypoint = table.remove(WptList, 1)
														if WptList[1] then
															Owner:RemoveMovePathBeginning()
														end
													else
														Waypoint = nil
													end
												end
												
												coroutine.yield()	-- wait until next frame
											elseif Owner.AIMode == Actor.AIMODE_GOLDDIG then
												Waypoint.Pos = SceneMan:MovePointToGround(Waypoint.Pos, Owner.Height*0.2, 4)
											end
										end
									end
								end
							else
								digState = AHuman.NOTDIGGING
								obstacleState = Actor.PROCEEDING
								AI.deviceState = AHuman.STILL
								AI.fire = false
								Owner:EquipFirearm(true)
							end
						else	-- not digging
							if not AI.Target then
								AI.fire = false
							end
							
							-- Scan for obstacles
							local Trace = Vector(Owner.Diameter*0.85, 0):RadRotate(scanAng)
							local Free = Vector()
							local index = math.floor(scanAng*2.5+2.01)
							if SceneMan:CastObstacleRay(Owner.Pos, Trace, Vector(), Free, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 3) > -1 then
								Obstacles[index] = true
							else
								Obstacles[index] = false
							end
							
							if scanAng < 1.57 then	-- pi/2
								if scanAng > 1.2 then
									scanAng = 1.89
								else
									scanAng = scanAng + 0.55
								end
							else
								if scanAng > 3.5 then
									scanAng = -0.4
								else
									scanAng = scanAng + 0.55
								end
							end
							
							if not AI.jump and not AI.flying then
								coroutine.yield()	-- wait until next frame
							end
							
							if CurrDist.Magnitude > Owner.Height * 0.4 then	-- not close enough to the waypoint
								ArrivedTimer:Reset()
								
								-- check if we have LOS to the waypoint
								if SceneMan:CastObstacleRay(Owner.Pos, CurrDist, Vector(), Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 9) < 0 then
									NoLOSTimer:Reset()
								elseif NoLOSTimer:IsPastSimTimeLimit() then	-- calulate new path
									Waypoint = nil
									WptList = nil -- update the path
									nextLatMove = Actor.LAT_STILL
									
									if Owner.AIMode == Actor.AIMODE_GOLDDIG and digState == AHuman.NOTDIGGING and math.random() < 0.5 then
										return true	-- end this behavior and look for gold again
									end
								end
								
								if not AI.jump and not AI.flying then
									coroutine.yield()	-- wait until next frame
								end
							elseif ArrivedTimer:IsPastSimTimeLimit() then	-- only remove a waypoint if we have been close to it for a while
								if Waypoint.Type == "last" then
									if not AI.flying and Owner.Vel.Largest < 5 then
										if not Owner.MOMoveTarget then
											local ProxyWpt = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height*0.2, 4)
											if SceneMan:ShortestDistance(Owner.Pos, ProxyWpt, false).Largest < Owner.Height*0.4 then
												Owner:ClearAIWaypoints()
												Owner:ClearMovePath()
												Owner:DrawWaypoints(false)
												break
											end
										end
										
										PrevWptPos = Waypoint.Pos
										Owner:RemoveMovePathBeginning()
										Waypoint = nil
										WptList = nil -- update the path
									end
								else
									PrevWptPos = Waypoint.Pos
									Owner:RemoveMovePathBeginning()
									Waypoint = nil
								end
							end
							
							if Waypoint then	-- move towards the waypoint
								-- control horisontal movement
								if Owner.FGLeg or Owner.BGLeg then
									if not AI.flying then
										if CurrDist.X < -3 then
											nextLatMove = Actor.LAT_LEFT
										elseif CurrDist.X > 3 then
											nextLatMove = Actor.LAT_RIGHT
										else
											nextLatMove = Actor.LAT_STILL
										end
									end
								elseif ((CurrDist.X < -5 and Owner.HFlipped) or (CurrDist.X > 5 and not Owner.HFlipped)) and math.abs(Owner.Vel.X) < 1 then
									-- no legs, jump forward
									AI.jump = true
								end
								
								if Waypoint.Type == "right" then
									if CurrDist.X > -3 then
										nextLatMove = Actor.LAT_RIGHT
									end
								elseif Waypoint.Type == "left" then
									if CurrDist.X < 3 then
										nextLatMove = Actor.LAT_LEFT
									end
								end
								
								if Owner.Jetpack and Owner.Head and Owner.Head:IsAttached() then
									if Owner.JetTimeLeft < AI.minBurstTime then
										AI.jump = false	-- not enough fuel left, no point in jumping yet
										if not AI.flying or Owner.Vel.Y > 1 then
											AI.refuel = true
										end
									else
										-- do we have a target we want to shoot at?
										if (AI.Target and AI.canHitTarget and AI.BehaviorName ~= "AttackTarget") then
											-- are we also flying
											if AI.flying then
												-- predict jetpack movement when jumping and there is a target (check one direction)
												local jetStrength = AI.jetImpulse / Owner.Mass
												local t = math.min(0.4, Owner.JetTimeLeft*0.001)
												local PixelVel = Owner.Vel * (FrameMan.PPM * t)
												local Accel = SceneMan.GlobalAcc * FrameMan.PPM
												
												-- a burst use 10x more fuel
												if Owner.Jetpack:CanTriggerBurst() then
													t = math.max(math.min(0.4, Owner.JetTimeLeft*0.001-TimerMan.DeltaTimeSecs*10), TimerMan.DeltaTimeSecs)
												end
												
												-- test jumping
												local JetAccel = Accel + Vector(-jetStrength, 0):RadRotate(Owner.RotAngle+1.375*math.pi+Owner:GetAimAngle(false)*0.25)
												local JumpPos = Owner.Head.Pos + PixelVel + JetAccel * (t*t*0.5)
												
												-- a burst add a one time boost to acceleration
												if Owner.Jetpack:CanTriggerBurst() then
													JumpPos = JumpPos + Vector(-AI.jetBurstAcc, 0):AbsRotateTo(JetAccel)
												end
												
												-- check for obstacles from the head
												Trace = SceneMan:ShortestDistance(Owner.Head.Pos, JumpPos, false)
												local jumpScore = SceneMan:CastObstacleRay(Owner.Head.Pos, Trace, JumpPos, Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 3)
												if jumpScore < 0 then	-- no obstacles: calculate the distance from the future pos to the wpt
													jumpScore = SceneMan:ShortestDistance(Waypoint.Pos, JumpPos, false).Magnitude
												else -- the ray hit terrain or start inside terrain: avoid
													jumpScore = SceneMan:ShortestDistance(Waypoint.Pos, JumpPos, false).Largest * 2
												end
												
												-- test falling
												local FallPos = Owner.Head.Pos + PixelVel + Accel * (t*t*0.5)
												
												-- check for obstacles when falling/walking
												local Trace = SceneMan:ShortestDistance(Owner.Head.Pos, FallPos, false)
												SceneMan:CastObstacleRay(Owner.Head.Pos, Trace, FallPos, Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 3)
												
												if jumpScore > SceneMan:ShortestDistance(Waypoint.Pos, FallPos, false).Magnitude then
													AI.jump = false
												else
													AI.jump = true
												end
											else
												AI.jump = false
											end
										else
											if Waypoint.Type ~= "drop" and not Lower(Waypoint, Owner, 20) then
												-- jump over low obstacles unless we want to jump off a ledge
												if nextLatMove == Actor.LAT_RIGHT and
													(Obstacles[Obst.R_LOW] or Obstacles[Obst.R_FRONT]) and not Obstacles[Obst.R_UP]
												then
													AI.jump = true
													if Obstacles[Obst.R_HIGH] then
														nextLatMove = Actor.LAT_LEFT -- TODO: only when too close to the obstacle?
													end
												elseif nextLatMove == Actor.LAT_LEFT and
													(Obstacles[Obst.L_LOW] or Obstacles[Obst.L_FRONT]) and not Obstacles[Obst.L_UP]
												then
													AI.jump = true
													if Obstacles[Obst.L_HIGH] then
														nextLatMove = Actor.LAT_RIGHT -- TODO: only when too close to the obstacle?
													end
												end
											end
											
											-- predict jetpack movement...
											local jetStrength = AI.jetImpulse / Owner.Mass
											local t = math.min(0.4, Owner.JetTimeLeft*0.001)
											local PixelVel = Owner.Vel * (FrameMan.PPM * t)
											local Accel = SceneMan.GlobalAcc * FrameMan.PPM
											
											-- a burst use 10x more fuel
											if Owner.Jetpack:CanTriggerBurst() then
												t = math.max(math.min(0.4, Owner.JetTimeLeft*0.001-TimerMan.DeltaTimeSecs*10), TimerMan.DeltaTimeSecs)
											end
											
											-- when jumping (check four directions)
											for k, Face in pairs(Facings) do
												local JetAccel = Vector(-jetStrength, 0):RadRotate(Owner.RotAngle+1.375*math.pi+Face.facing*0.25)
												local JumpPos = Owner.Head.Pos + PixelVel + (Accel + JetAccel) * (t*t*0.5)
												
												-- a burst add a one time boost to acceleration
												if Owner.Jetpack:CanTriggerBurst() then
													JumpPos = JumpPos + Vector(-AI.jetBurstAcc, 0):AbsRotateTo(JetAccel)
												end
												
												-- check for obstacles from the head
												Trace = SceneMan:ShortestDistance(Owner.Head.Pos, JumpPos, false)
												local obstDist = SceneMan:CastObstacleRay(Owner.Head.Pos, Trace, JumpPos, Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 3)
												if obstDist < 0 then	-- no obstacles: calculate the distance from the future pos to the wpt
													Facings[k].range = SceneMan:ShortestDistance(Waypoint.Pos, JumpPos, false).Magnitude
												else -- the ray hit terrain or start inside terrain: avoid
													Facings[k].range = SceneMan:ShortestDistance(Waypoint.Pos, JumpPos, false).Largest * 2
												end
											end
											
											-- when falling or walking
											local FallPos = Owner.Head.Pos + PixelVel
											if AI.flying then
												FallPos = FallPos + Accel * (t*t*0.5)
											end
											
											-- check for obstacles when falling/walking
											local Trace = SceneMan:ShortestDistance(Owner.Head.Pos, FallPos, false)
											SceneMan:CastObstacleRay(Owner.Head.Pos, Trace, FallPos, Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 3)
											
											table.sort(Facings, function(A, B) return A.range < B.range end)
											local delta = SceneMan:ShortestDistance(Waypoint.Pos, FallPos, false).Magnitude - Facings[1].range
											if delta < 1 then
												AI.jump = false
											elseif AI.flying or delta > 15 then
												AI.jump = true
												nextAimAngle = Owner:GetAimAngle(false) * 0.5 + Facings[1].aim * 0.5	-- adjust jetpack nozzle direction
												nextLatMove = Actor.LAT_STILL
												
												if Facings[1].facing > 1.4 then
													if not Owner.HFlipped then
														nextLatMove = Actor.LAT_LEFT
													end
												elseif Owner.HFlipped then
													nextLatMove = Actor.LAT_RIGHT
												end
											end
										end
									end
								end
							end
						end
					end
				end
			end
		else	-- no waypoint list
			local Trace = SceneMan:ShortestDistance(Owner.Pos, Owner:GetLastAIWaypoint(), false)
			
			-- if digging for gold: ignore the path-finding and plot waypoints in a straight line to the target unless there is metal in the way
			if Owner.AIMode == Actor.AIMODE_GOLDDIG and not AI.Target and Owner:EquipDiggingTool(false) and not SceneMan:CastStrengthRay(Owner.Pos, Trace, 105, Vector(), 1, rte.grassID, true) then
				WptList = {}	-- store the waypoints we want in our path here
				
				local wpts = math.ceil(Trace.Magnitude/60)
				Trace:CapMagnitude(60)
				for i = 1, wpts do
					local TmpPos = Owner.Pos + Trace * i
					table.insert(WptList, {Pos=SceneMan:MovePointToGround(TmpPos, Owner.Height*0.2, 4)})
				end
				
				if WptList[wpts] then
					WptList[wpts].Type = "last"
					
					-- create the move path seen on the screen
					for _, Wpt in pairs(WptList) do
						Owner:AddToMovePathEnd(Wpt.Pos)
					end
					
					Owner:DrawWaypoints(true)
					NoLOSTimer:Reset()
				end
			else
				Owner:UpdateMovePath()
				
				-- have we arrived?
				if not Owner.MOMoveTarget then
					local ProxyWpt = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height*0.2, 5)
					local Trace = SceneMan:ShortestDistance(Owner.Pos, ProxyWpt, false)
					if Trace.Largest < Owner.Height*0.25 and not SceneMan:CastStrengthRay(Owner.Pos, Trace, 6, Vector(), 3, rte.grassID, true) then
						if Owner.AIMode == Actor.AIMODE_GOTO then
							AI.SentryPos = Vector(Owner.Pos.X, Owner.Pos.Y)
							AI:CreateSentryBehavior(Owner)
						end
						
						Owner:ClearAIWaypoints()
						Owner:ClearMovePath()
						Owner:DrawWaypoints(false)
						
						break
					end
				end
				
				-- no waypoint list, create one in several small steps to reduce lag
				local PathDump = {}
				if Owner.MOMoveTarget and MovableMan:ValidMO(Owner.MOMoveTarget) then
					Owner:DrawWaypoints(false)
					
					-- copy the MovePath to a temporary table so we can yield safely while working on the path
					for WptPos in Owner.MovePath do
						table.insert(PathDump, WptPos)
					end
					
					-- clear the path here so the graphical representation does not flicker on and off as much
					Owner:ClearMovePath()
					Owner:AddToMovePathEnd(Owner.MOMoveTarget.Pos)
				else
					coroutine.yield()	-- wait until next frame
					
					-- copy the MovePath to a temporary table so we can yield safely while working on the path
					for WptPos in Owner.MovePath do
						table.insert(PathDump, WptPos)
					end
				end
				
				coroutine.yield()	-- wait until next frame
				
				-- copy useful waypoints to a temporary path
				local TmpWpts = {}
				table.insert(TmpWpts, {Pos=Owner.Pos})
				local Origin
				local LastPos = PathDump[1]
				local index = 1
				for _, WptPos in pairs(PathDump) do
					Origin = TmpWpts[index].Pos
					local Dist = SceneMan:ShortestDistance(Origin, WptPos, false)
					if math.abs(Dist.Y) > 30 or Dist.Magnitude > 80 or	-- skip any waypoint too close to the previous one
						SceneMan:CastStrengthSumRay(Origin, WptPos, 3, rte.grassID) > 5
					then
						table.insert(TmpWpts, {Pos=LastPos})
						index = index + 1
					end
					
					LastPos = WptPos
					coroutine.yield()	-- wait until next frame
				end
				
				table.insert(TmpWpts, {Pos=PathDump[#PathDump]})	-- add the last waypoint in the MovePath
				
				WptList = {}	-- store the waypoints we want in our path here
				local StartWpt = table.remove(TmpWpts, 1)
				while TmpWpts[1] do
					local NextWpt = table.remove(TmpWpts, 1)
					
					if Lower(NextWpt, StartWpt, 30) then	-- scan for sharp drops
						local Dist = SceneMan:ShortestDistance(StartWpt.Pos, NextWpt.Pos, false)
						if math.abs(Dist.X) < Dist.Y then -- check the slope
							if SceneMan:CastObstacleRay(StartWpt.Pos, Dist, Vector(), Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 4) < 0 then
								NextWpt.Type = "drop"	-- LOS from StartWpt to NextWpt
							end
							
							local GapList = {}
							for j, JumpWpt in pairs(TmpWpts) do	-- look for the other side
								local Gap = SceneMan:ShortestDistance(StartWpt.Pos, JumpWpt.Pos, false)
								if Gap.Magnitude > 400 - Gap.Y then	-- TODO: use actor properties here
									break	-- too far
								end
								
								if Gap.Y > -40 then	-- no more than 2m above
									table.insert(GapList, {Wpt=JumpWpt, score=math.abs(Gap.X/Gap.Y), index=j})
								end
							end
							
							coroutine.yield()	-- wait until next frame
							
							table.sort(GapList, function(A, B) return A.score > B.score end)	-- sort largest first
							
							for _, LZ in pairs(GapList) do
								-- check if we can jump
								local Trace = SceneMan:ShortestDistance(StartWpt.Pos, LZ.Wpt.Pos, false)
								if SceneMan:CastObstacleRay(StartWpt.Pos, Trace, Vector(), Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 4) < 0 then
									-- find a point mid-air
									local TestPos = StartWpt.Pos + Trace * 0.6
									local Free = Vector()
									if 0 ~= SceneMan:CastObstacleRay(TestPos, Vector(0, -math.abs(Trace.X)/2), Vector(), Free, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 2) then	-- TODO: check LOS? what if 0?
										table.insert(WptList, {Pos=Free+Vector(0,Owner.Height/3), Type="air"})	-- guide point in the air
										NextWpt = LZ.Wpt
										
										-- delete any waypoints between StartWpt and the LZ
										for i = LZ.index, 1, -1 do
											table.remove(TmpWpts, i)
										end
										
										break
									end
								end
								
								coroutine.yield()	-- wait until next frame
							end
						end
					end
					
					table.insert(WptList, NextWpt)
					StartWpt = NextWpt
				end
				
				WptList[#WptList].Type = "last"
				
				if not Owner.MOMoveTarget then
					Owner:ClearMovePath()
					for _, Wpt in pairs(WptList) do
						Owner:AddToMovePathEnd(Wpt.Pos)
					end
				end
				
				Owner:DrawWaypoints(true)
				NoLOSTimer:Reset()
			end
		end
		
		-- movement commands
		if (AI.Target and AI.BehaviorName ~= "AttackTarget") or
			(Owner.AIMode ~= Actor.AIMODE_SQUAD and (AI.BehaviorName == "ShootArea" or AI.BehaviorName == "FaceAlarm"))
		then
			AI.lateralMoveState = Actor.LAT_STILL
			if not AI.flying then
				AI.jump = false
			end
		else
			AI.lateralMoveState = nextLatMove
			if digState == AHuman.NOTDIGGING then
				Owner:SetAimAngle(nextAimAngle)
				nextAimAngle = Owner:GetAimAngle(false) * 0.95	-- look straight ahead
			end
		end
		
		if AI.BlockingMO then
			if not MovableMan:ValidMO(AI.BlockingMO) or
				SceneMan:ShortestDistance(Owner.Pos, AI.BlockingMO.Pos, false).Largest > (Owner.Height + AI.BlockingMO.Diameter)*1.2
			then
				AI.BlockingMO = nil
				AI.teamBlockState = Actor.NOTBLOCKED
			elseif AI.teamBlockState == Actor.NOTBLOCKED and Waypoint then
				if (Waypoint.Pos.X > Owner.Pos.X and AI.BlockingMO.Pos.X > Owner.Pos.X) or
					 (Waypoint.Pos.X < Owner.Pos.X and AI.BlockingMO.Pos.X < Owner.Pos.X)
				then
					AI.teamBlockState = Actor.BLOCKED
				else
					AI.BlockingMO = nil
				end
			end
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end


-- go prone if we can shoot from the prone position and return the result
function MeleeHumanBehaviors.GoProne(AI, Owner, TargetPos, targetID)
	if not Owner.Head or AI.proneState == AHuman.PRONE then
		return false
	end
	
	-- only go prone if we can see the ground near the target
	local AimPoint = SceneMan:MovePointToGround(TargetPos, 10, 3)
	local ground = Owner.Pos.Y + Owner.Height * 0.25
	local Dist = SceneMan:ShortestDistance(Owner.Pos, AimPoint, false)
	local PronePos
	
	-- check if there is room to go prone here
	if Dist.X > Owner.Height then
		-- to the right
		PronePos = Owner.EyePos + Vector(Owner.Height*0.3, 0)
		
		local x_pos = Owner.Pos.X + 10
		for _ = 1, math.ceil(Owner.Height/16) do
			x_pos = x_pos + 7
			if SceneMan.SceneWrapsX and x_pos >= SceneMan.SceneWidth then
				x_pos = SceneMan.SceneWidth - x_pos
			end
			
			if 0 == SceneMan:GetTerrMatter(x_pos, ground) then
				return false
			end
		end
	elseif Dist.X < -Owner.Height then
		-- to the left
		PronePos = Owner.EyePos + Vector(-Owner.Height*0.3, 0)
		
		local x_pos = Owner.Pos.X - 10
		for _ = 1, math.ceil(Owner.Height/16) do
			x_pos = x_pos - 7
			if SceneMan.SceneWrapsX and x_pos < 0 then
				x_pos = x_pos + SceneMan.SceneWidth
			end
			
			if 0 == SceneMan:GetTerrMatter(x_pos, ground) then
				return false
			end
		end
	else
		return false	-- target is too close
	end
	
	PronePos = SceneMan:MovePointToGround(PronePos, Owner.Head.Radius+3, 2)
	Dist = SceneMan:ShortestDistance(PronePos, AimPoint, false)
	
	-- check LOS from the prone position
	--if not SceneMan:CastFindMORay(PronePos, Dist, targetID, Hit, rte.grassID, false, 8) then
	if SceneMan:CastObstacleRay(PronePos, Dist, Vector(), Vector(), targetID, Owner.IgnoresWhichTeam, rte.grassID, 9) > -1 then
		return false
	else
		-- check for obstacles more more carefully
		Dist:CapMagnitude(60)
		if SceneMan:CastObstacleRay(PronePos, Dist, Vector(), Vector(), 0, Owner.IgnoresWhichTeam, rte.grassID, 1) > -1 then
			return false
		end
	end
	
	AI.proneState = AHuman.PRONE
	if Dist.X > 0 then
		AI.lateralMoveState = Actor.LAT_RIGHT
	else
		AI.lateralMoveState = Actor.LAT_LEFT
	end
	
	return true
end

-- get the projectile properties from the magazine
function MeleeHumanBehaviors.GetProjectileData(Owner)
	local Weapon = ToHDFirearm(Owner.EquippedItem)
	local Round = Weapon.Magazine.NextRound
	local Projectile = Round.NextParticle
	local PrjDat = {MagazineName=Weapon.Magazine.PresetName}
	
	if Round.IsEmpty then	-- set default values if there is no particle
		PrjDat.g = 0
		PrjDat.vel = 100
		PrjDat.rng = math.huge
	else
		PrjDat.blast = Weapon:GetAIBlastRadius() -- check if this weapon have a blast radius
		if PrjDat.blast > 0 then
			PrjDat.exp = true	-- set this for legacy reasons
		end
		
		-- find muzzle velocity
		PrjDat.vel = Weapon:GetAIFireVel()
		
		-- half of the theoretical upper limit for the total amount of material strength this weapon can destroy in 250ms
		PrjDat.pen = Weapon:GetAIPenetration() * math.max((Weapon.RateOfFire / 240), 1)
		
		PrjDat.g = SceneMan.GlobalAcc.Y * 0.67 * Weapon:GetBulletAccScalar()	-- underestimate gravity
		PrjDat.vsq = PrjDat.vel^2	-- muzzle velocity squared
		PrjDat.vqu = PrjDat.vsq^2	-- muzzle velocity quadrat
		PrjDat.drg = 1 - Projectile.AirResistance * TimerMan.DeltaTimeSecs	-- AirResistance is stored as the ini-value times 60
		PrjDat.thr = math.min(Projectile.AirThreshold, PrjDat.vel)
		
		-- estimate theoretical max range with ...
		local lifeTime = Weapon:GetAIBulletLifeTime()
		if lifeTime < 1 then	-- infinite life time
			PrjDat.rng = math.huge
		elseif PrjDat.drg < 1 then	-- AirResistance
			PrjDat.rng = 0
			local threshold = PrjDat.thr * FrameMan.PPM * TimerMan.DeltaTimeSecs	-- AirThreshold in pixels/frame
			local vel = PrjDat.vel * FrameMan.PPM * TimerMan.DeltaTimeSecs	-- muzzle velocity in pixels/frame
			for _ = 0, math.ceil(lifeTime/TimerMan.DeltaTimeMS) do
				PrjDat.rng = PrjDat.rng + vel
				if vel > threshold then
					vel = vel * PrjDat.drg
				end
			end
		else	-- no AirResistance
			PrjDat.rng = PrjDat.vel * FrameMan.PPM * TimerMan.DeltaTimeSecs * (lifeTime / TimerMan.DeltaTimeMS)
		end
	end
	
	return PrjDat
end

-- open fire on the selected target
function MeleeHumanBehaviors.ShootTarget(AI, Owner)
	if not MovableMan:ValidMO(AI.Target) then
		return true
	end

	AI.canHitTarget = false
	AI.TargetLostTimer:SetSimTimeLimitMS(700)
	
	local LOSTimer = Timer()
	LOSTimer:SetSimTimeLimitMS(170)
	
	local ShootTimer = Timer()
	local shootDelay = RangeRand(440, 590) * AI.aimSpeed + 150
	local AimPoint = AI.Target.Pos + AI.TargetOffset
	if not AI.flying and AI.Target.Vel.Largest < 4 and MeleeHumanBehaviors.GoProne(AI, Owner, AimPoint, AI.Target.ID) then
		shootDelay = shootDelay + 250
	end
	
	-- spin up asap
	if Owner.FirearmActivationDelay > 0 then
		shootDelay = math.max(50*AI.aimSpeed, shootDelay-Owner.FirearmActivationDelay)
	end
	
	local PrjDat
	local openFire = 0
	local checkAim = true
	local TargetAvgVel = Vector(AI.Target.Vel.X, AI.Target.Vel.Y)
	local Dist = SceneMan:ShortestDistance(Owner.Pos, AimPoint, false)
	
	-- make sure we are facing the right direction
	if Owner.HFlipped then
		if Dist.X > 0 then
			Owner.HFlipped = false
		end
	elseif Dist.X < 0 then
		Owner.HFlipped = true
	end
	
	-- alert nearby allies  TODO: do this better engine-side
	for i = 0.95, 0.5, -0.4 do
		local Alert = CreateTDExplosive("Alert Device "..math.random(3), "Base.rte")
		Alert.Pos = Owner.Pos + Dist * i
		Alert.Team = AI.Target.Team
		Alert:Activate()
		MovableMan:AddParticle(Alert)
	end
	
	coroutine.yield()	-- wait until next frame
	
	local distMultiplier = AI.aimSkill * math.max(math.min(0.0035*Dist.Largest, 1.0), 0.01)
	local ErrorOffset = Vector(RangeRand(60, 110)*distMultiplier, 0):RadRotate(RangeRand(1, 6))
	local aimTarget = SceneMan:ShortestDistance(Owner.Pos, AimPoint+ErrorOffset, false).AbsRadAngle
	local f1, f2 = 0.5, 0.5 -- aim noise filter
	
	while true do
		if not AI.Target or AI.Target:IsDead() then
			AI.Target = nil
			
			-- the target is gone, try to find another right away
			local ClosestEnemy = MovableMan:GetClosestEnemyActor(Owner.Team, AimPoint, 200, Vector())
			if ClosestEnemy and not ClosestEnemy:IsDead() then
				if ClosestEnemy.ClassName == "AHuman" then
					ClosestEnemy = ToAHuman(ClosestEnemy)
				elseif ClosestEnemy.ClassName == "ACrab" then
					ClosestEnemy = ToACrab(ClosestEnemy)
				else
					ClosestEnemy = nil
				end
				
				if ClosestEnemy then
					-- check if the target is inside our "screen"
					local ViewDist = SceneMan:ShortestDistance(Owner.ViewPoint, ClosestEnemy.Pos, false)
					if (math.abs(ViewDist.X) - ClosestEnemy.Radius < FrameMan.PlayerScreenWidth * 0.5) and
						(math.abs(ViewDist.Y) - ClosestEnemy.Radius < FrameMan.PlayerScreenHeight * 0.5)
					then
						if not AI.isPlayerOwned or not SceneMan:IsUnseen(ClosestEnemy.Pos.X, ClosestEnemy.Pos.Y, Owner.Team) then	-- AI-teams ignore the fog
							if SceneMan:CastStrengthSumRay(Owner.EyePos, ClosestEnemy.Pos, 6, rte.grassID) < 120 then
								AI.Target = ClosestEnemy
								AI.TargetOffset = Vector()
								AimPoint = AimPoint * 0.3 + AI.Target.Pos * 0.7
							end
						end
					end
				end
			end
			
			-- no new target found
			if not AI.Target then
				break
			end
		end
		
		if Owner.FirearmIsReady then
			-- it is now safe to get the ammo stats since FirearmIsReady
			local Weapon = ToHDFirearm(Owner.EquippedItem)
			if not PrjDat or PrjDat.MagazineName ~= Weapon.Magazine.PresetName then
				PrjDat = MeleeHumanBehaviors.GetProjectileData(Owner)
				
				-- uncomment these to get the range of the weapon
				--ConsoleMan:PrintString(Weapon.PresetName .. " range = " .. PrjDat.rng .. " px")
				--ConsoleMan:PrintString(AI.Target.PresetName .. " range = " .. SceneMan:ShortestDistance(Owner.Pos, AI.Target.Pos, false).Magnitude .. " px")
				
				-- Aim longer with low capacity weapons
				if Weapon.Magazine.Capacity > -1 and Weapon.Magazine.Capacity < 10 and Dist.Largest > 150 then
					ErrorOffset = ErrorOffset * 0.6
					shootDelay = shootDelay + 500
				end
			else
				-- TODO: make low skill AI lead worse
				TargetAvgVel = TargetAvgVel * 0.8 + AI.Target.Vel * 0.2	-- smooth the target's velocity
				AimPoint = AI.Target.Pos + AI.TargetOffset + ErrorOffset
				
				Dist = SceneMan:ShortestDistance(Weapon.Pos, AimPoint, false)
				local range = Dist.Magnitude
				
				if range < 100 then
					-- move the aimpoint towards the center of the target at close ranges
					if range < 50 then
						AI.TargetOffset = AI.TargetOffset * 0.95
						TargetAvgVel:Reset()	-- high vel at close range confuse the AI tracking
					else
						TargetAvgVel = TargetAvgVel * 0.5
					end
				end
				
				if checkAim then
					checkAim = false	-- only check every second frame
					
					if range < PrjDat.blast then
						-- it is not safe to fire an explosive projectile at this distance
						aimTarget = Dist.AbsRadAngle
						AI.canHitTarget = true
						if Owner.InventorySize > 0 then	-- we have more things in the inventory
							if range < 60 and Owner:HasObjectInGroup("Diggers") then
								AI:CreateHtHBehavior(Owner)
								break
							elseif Owner:EquipLoadedFirearmInGroup("Any", "Explosive Weapons", true) then
								PrjDat = nil
								coroutine.yield()
							else
								PrjDat.blast = 0 -- no other loaded weapon, ignore the blast radius
							end
						else
							PrjDat.blast = 0 -- no other weapon, ignore the blast radius
						end
					elseif range < PrjDat.rng then
						-- lead the target if target speed and projectile TTT is above the threshold
						local timeToTarget = range / PrjDat.vel
						if timeToTarget * AI.Target.Vel.Magnitude > 2 then
							-- ~double this value since we only do this every second update
							if PosRand() > 0.5 then
								timeToTarget = timeToTarget * (2 - RangeRand(0, 0.4) * AI.aimSkill)
							else
								timeToTarget = timeToTarget * (2 + RangeRand(0, 0.4) * AI.aimSkill)
							end
							
							Dist = SceneMan:ShortestDistance(Weapon.Pos, AimPoint+(Owner.Vel*0.5+AI.Target.Vel)*timeToTarget, false)
						end
						
						aimTarget = MeleeHumanBehaviors.GetAngleToHit(PrjDat, Dist)
						if aimTarget then
							AI.canHitTarget = true
						else
							AI.canHitTarget = false
							
							-- the target is too far away
							if not AI.isPlayerOwned or Owner.AIMode ~= Actor.AIMODE_SENTRY then
								if not Owner.MOMoveTarget or 
									not MovableMan:ValidMO(Owner.MOMoveTarget) or
									Owner.MOMoveTarget.RootID ~= AI.Target.RootID
								then	-- move towards the target
									local OldWaypoint = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height/5, 4)	-- move back here later
									Owner:ClearAIWaypoints()
									Owner:AddAIMOWaypoint(AI.Target)
									Owner:AddAISceneWaypoint(OldWaypoint)
									AI:CreateGoToBehavior(Owner)
									AI.proneState = AHuman.NOTPRONE
								end
							else
								-- TODO: switch weapon properly
								if Weapon:HasObjectInGroup("Primary Weapons") then
									if Owner:EquipDeviceInGroup("Secondary Weapons", true) then
										coroutine.yield()
										PrjDat = nil
									end
								elseif Weapon:HasObjectInGroup("Secondary Weapons") then
									if Owner:EquipDeviceInGroup("Primary Weapons", true) then
										coroutine.yield()
										PrjDat = nil
									end
								end
							end
						end
					elseif not AI.isPlayerOwned or not (Owner.AIMode == Actor.AIMODE_SENTRY or Owner.AIMode == Actor.AIMODE_SQUAD) then -- target out of reach; move towards it
						-- check if we are already moving towards an actor
						if not Owner.MOMoveTarget or 
							not MovableMan:ValidMO(Owner.MOMoveTarget) or
							Owner.MOMoveTarget.RootID ~= AI.Target.RootID
						then	-- move towards the target							
							local OldWaypoint = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height/5, 4)	-- move back here later
							Owner:ClearAIWaypoints()
							Owner:AddAIMOWaypoint(AI.Target)
							Owner:AddAISceneWaypoint(OldWaypoint)
							AI:CreateGoToBehavior(Owner)
							AI.proneState = AHuman.NOTPRONE
							AI.canHitTarget = false
						end
					end
				else
					checkAim = true
					
					-- periodically check that we have LOS to the target
					if LOSTimer:IsPastSimTimeLimit() then
						LOSTimer:Reset()
						local TargetPoint = AI.Target.Pos + AI.TargetOffset
						
						if (range < 30 + Weapon.SharpLength + FrameMan.PlayerScreenWidth*0.5) and
							(not AI.isPlayerOwned or not SceneMan:IsUnseen(TargetPoint.X, TargetPoint.Y, Owner.Team))
						then
							if PrjDat.pen > 0 then
								if SceneMan:CastStrengthSumRay(Weapon.Pos, TargetPoint, 6, rte.grassID) * 5 < PrjDat.pen then
									AI.TargetLostTimer:Reset()	-- we can shoot at the target
									AI.OldTargetPos = Vector(AI.Target.Pos.X, AI.Target.Pos.Y)
								end
							else
								if SceneMan:CastStrengthSumRay(Weapon.Pos, TargetPoint, 6, rte.grassID) < 120 then
									AI.TargetLostTimer:Reset()	-- we can shoot at the target
									AI.OldTargetPos = Vector(AI.Target.Pos.X, AI.Target.Pos.Y)
								end
							end
						end
					end
				end
				
				if AI.canHitTarget then
					AI.lateralMoveState = Actor.LAT_STILL
					if not AI.flying then
						AI.deviceState = AHuman.AIMING
					end
				end
				
				-- add some filtered noise to the aim
				local aim = Owner:GetAimAngle(true)			
				local noise = RangeRand(-30, 30) * AI.aimSpeed
				f1, f2 = 0.9*f1+noise*0.1, 0.7*f2+noise*0.3
				noise = f1 + f2 + noise * 0.1
				aimTarget = (aimTarget or aim) + math.min(math.max(noise/(range+30), -0.12), 0.12)
				
				if AI.flying then
					aimTarget = aimTarget + RangeRand(-0.05, 0.05)
				end
				
				local angDiff = aim - aimTarget
				if angDiff > math.pi then
					angDiff = angDiff - math.pi*2
				elseif angDiff < -math.pi then
					angDiff = angDiff + math.pi*2
				end
				
				AI.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim-math.max(math.min(angDiff*(0.1/AI.aimSkill), 0.17/AI.aimSkill), -0.17/AI.aimSkill))
				
				if PrjDat and ShootTimer:IsPastRealMS(shootDelay) then
					ErrorOffset = ErrorOffset * 0.98	-- reduce the aim point error
					if AI.canHitTarget and angDiff < 0.7 then
						if Weapon.FullAuto then	-- open fire if our aim overlap the target
							if math.abs(angDiff) < math.tanh((AI.Target.Diameter*AI.aimSkill*2)/(range+1)) then
								openFire = 5	-- don't stop shooting just because we lose the target for a few frames
							else
								openFire = openFire - 1
							end
						elseif not AI.fire then	-- open fire if our aim overlap the target
							if math.abs(angDiff) < math.tanh((AI.Target.Diameter*AI.aimSkill*1.5)/(range+1)) then
								openFire = 1
							else
								openFire = 0
							end
						else
							openFire = openFire - 1	-- release the trigger if semi auto
						end
						
						-- check for obstacles if the ammo have a blast radius
						if openFire > 0 and PrjDat.blast > 0 then
							if SceneMan:CastObstacleRay(Weapon.MuzzlePos, Weapon:RotateOffset(Vector(50, 0)), Vector(), Vector(), Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, 2) > -1 then
								-- equip another primary if possible
								if Owner:EquipLoadedFirearmInGroup("Any", "Explosive Weapons", true) then
									PrjDat = nil
									openFire = 0
									coroutine.yield()
								else
									PrjDat.blast = 0
								end
							end
						end
					else
						openFire = openFire - 1
					end
				else
					ErrorOffset = ErrorOffset * 0.995	-- reduce the aim point error
					openFire = 0
				end
				
				if openFire > 0 then
					AI.fire = true
				else
					AI.fire = false
				end
			end
		else
			if Owner.EquippedItem and ToHeldDevice(Owner.EquippedItem):IsReloading() then
				ShootTimer:Reset()
				AI.Ctrl.AnalogAim = SceneMan:ShortestDistance(Owner.Pos, AI.Target.Pos, false).Normalized
			elseif Owner:EquipFirearm(true) then
				coroutine.yield()	-- wait until next frame, just in case the magazine is replentishied by another script
				
				if Owner.FirearmIsEmpty then
					AI.deviceState = AHuman.POINTING
					AI.fire = false
					openFire = 0
					
					if AI.Target then
						-- equip another primary if possible
						if Owner:EquipLoadedFirearmInGroup("Primary Weapons", "None", true) then
							PrjDat = nil
						else
							-- select a sidearm instead of reloading if the target is within half a screen
							if Dist.Largest < (FrameMan.PlayerScreenWidth * 0.5 + AI.Target.Radius + Owner.AimDistance) then
								-- select a primary if we have an empty secondary equipped
								if Owner:EquipLoadedFirearmInGroup("Secondary Weapons", "None", true) then
									PrjDat = nil
								elseif Owner:EquipDeviceInGroup("Primary Weapons", true) then
									PrjDat = nil
								end
							end
						end
						
						coroutine.yield()
					end
					
					-- we might have a different weapon equipped now check if FirearmIsEmpty again
					if Owner.FirearmIsEmpty then
						Owner:ReloadFirearm()
						-- TODO: check if ducking is appropriate while reloading (when we can make the actor stand up reliably)
					end
					
					local distMultiplier = AI.aimSkill * math.max(math.min(0.006*Dist.Largest+0.1, 1.5), 0.2)
					ErrorOffset = ErrorOffset + Vector(RangeRand(5, 15)*distMultiplier, 0):RadRotate(RangeRand(1, 6))
					shootDelay = RangeRand(220, 330) * AI.aimSpeed + 50
					if Owner.FirearmActivationDelay > 0 then	-- Spin up asap
						shootDelay = math.max(50*AI.aimSpeed, shootDelay-Owner.FirearmActivationDelay)
					end
				end
			else
				AI:CreateGetWeaponBehavior(Owner)
				break -- no firearm avaliable
			end
		end
		
		-- make sure we are facing the right direction
		if Owner.HFlipped then
			if Dist.X > 0 then
				Owner.HFlipped = false
			end
		elseif Dist.X < 0 then
			Owner.HFlipped = true
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	-- reload the sidearm before switching to the primary weapon
	if Owner:HasObjectInGroup("Primary Weapons") and (Owner.EquippedItem and Owner.EquippedItem:HasObjectInGroup("Secondary Weapons")) then
		while Owner.EquippedItem and ToHeldDevice(Owner.EquippedItem):IsReloading() do
			coroutine.yield()	-- wait here until the current weapon has finished reloading
		end
	end
	
	if AI.PlayerPreferredHD then
		Owner:EquipNamedDevice(AI.PlayerPreferredHD, true)
	elseif not Owner:EquipDeviceInGroup("Primary Weapons", true) then
		Owner:EquipDeviceInGroup("Secondary Weapons", true)
	end
	
	if Owner.FirearmIsEmpty then
		Owner:ReloadFirearm()
	end
	
	return true
end

-- throw a grenade at the selected target
function MeleeHumanBehaviors.ThrowTarget(AI, Owner)
	local ThrowTimer = Timer()
	local aimTime = RangeRand(1000, 1500)
	local scan = 0
	local miss = 0	-- stop scanning after a few missed atempts
	local AimPoint, Dist, MO, ID, rootID, aim, LOS
	
	AI.TargetLostTimer:SetSimTimeLimitMS(1500)
	
	while true do
		if not MovableMan:ValidMO(AI.Target) then
			break
		end
		
		if scan < 1 then
			if AI.Target.Door then
				AimPoint = AI.Target.Door.Pos
			else
				AimPoint = AI.Target.Pos	-- look for the center
				if AI.Target.EyePos then
					AimPoint = (AimPoint + AI.Target.EyePos) / 2
				end
			end
			
			ID = rte.NoMOID
			if Owner:IsWithinRange(Vector(AimPoint.X, AimPoint.Y)) then	-- TODO: use grenade properies to decide this
				Dist = SceneMan:ShortestDistance(Owner.EyePos, AimPoint, false)
				ID = SceneMan:CastMORay(Owner.EyePos, Dist, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, false, 3)
				if ID < 1 or ID > 254 then	-- not found, look for any head or legs
					AimPoint = AI.Target.EyePos	-- the head
					if AimPoint then
						coroutine.yield()	-- wait until next frame
						if not MovableMan:ValidMO(AI.Target) then	-- must verify that the target exist after a yield
							break
						end
						
						Dist = SceneMan:ShortestDistance(Owner.EyePos, AimPoint, false)
						ID = SceneMan:CastMORay(Owner.EyePos, Dist, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, false, 3)
					end
					
					if ID < 1 or ID > 254 then
						local Legs = AI.Target.FGLeg or AI.Target.BGLeg	-- the legs
						if Legs then
							coroutine.yield()	-- wait until next frame
							if not MovableMan:ValidMO(AI.Target) then	-- must verify that the target exist after a yield
								break
							end
							
							AimPoint = Legs.Pos
							Dist = SceneMan:ShortestDistance(Owner.EyePos, AimPoint, false)
							ID = SceneMan:CastMORay(Owner.EyePos, Dist, Owner.ID, Owner.IgnoresWhichTeam, rte.grassID, false, 3)
						end
					end
				end
			else
				break	-- out of range
			end
			
			if ID > 0 and ID < rte.NoMOID then	-- MO found
				scan = 6	-- skip the LOS check the next n frames
				miss = 0
				LOS = true	-- we have line of sight to the target
				
				-- check what target we will hit
				rootID = MovableMan:GetRootMOID(ID)
				if rootID ~= AI.Target.ID then
					MO = MovableMan:GetMOFromID(rootID)
					if MovableMan:ValidMO(MO) then
						if MO.Team ~= Owner.Team then
							if MO.ClassName == "AHuman" then
								AI.Target = ToAHuman(MO)
							elseif MO.ClassName == "ACrab" then
								AI.Target = ToACrab(MO)
							elseif MO.ClassName == "ACRocket" then
								AI.Target = ToACRocket(MO)
							elseif MO.ClassName == "ACDropShip" then
								AI.Target = ToACDropShip(MO)
							elseif MO.ClassName == "ADoor" then
								AI.Target = ToADoor(MO)
							elseif MO.ClassName == "Actor" then
								AI.Target = ToActor(MO)
							else
								break
							end
						else
							break	-- don't shoot friendlies
						end
					end
				end
			else
				miss = miss + 1
				if miss > 4 then	-- stop looking if we cannot find anything after n atempts
					break
				else
					scan = 3	-- check LOS a little bit more often if no MO was found
				end
			end
			
			if LOS then	-- don't sharp aim until LOS has been confirmed
				if Owner.ThrowableIsReady then
					aim = MeleeHumanBehaviors.GetGrenadeAngle(AimPoint, AI.Target.Vel/RangeRand(0.5, 1.75), ToHeldDevice(Owner.EquippedItem).MuzzlePos, 18)
					if aim then
						AI.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim+RangeRand(-0.1, 0.1))
					else
						break	-- target out of range
					end
					
					if not ThrowTimer:IsPastSimMS(aimTime) then
						AI.fire = true
					else
						ThrowTimer:Reset()
						AI.fire = false
						aimTime = RangeRand(1000, 1500)
					end
				else
					break	-- no grenades left
				end
			end
		else
			scan = scan - 1
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	return true
end

-- attack the target in hand-to-hand
function MeleeHumanBehaviors.AttackTarget(AI, Owner)
	if not AI.Target or not MovableMan:ValidMO(AI.Target) then
		return true
	end
	
	AI.TargetLostTimer:SetSimTimeLimitMS(5000)
	
	-- move back here later
	local PrevMOMoveTarget, PrevSceneWaypoint
	if Owner.MOMoveTarget and MovableMan:ValidMO(Owner.MOMoveTarget) then
		PrevMOMoveTarget = Owner.MOMoveTarget
	else
		Owner.MOMoveTarget = nil
		PrevSceneWaypoint = SceneMan:MovePointToGround(Owner:GetLastAIWaypoint(), Owner.Height/5, 4)
	end
	
	-- move towards the target
	Owner:ClearMovePath()
	Owner:AddAIMOWaypoint(AI.Target)
	
	if PrevMOMoveTarget then
		Owner:AddAIMOWaypoint(PrevMOMoveTarget)
	end
	
	if PrevSceneWaypoint then
		Owner:AddAISceneWaypoint(PrevSceneWaypoint)
	end
	
	AI:CreateGoToBehavior(Owner)
	AI.proneState = AHuman.NOTPRONE
	
	while true do
		coroutine.yield()	-- wait until next frame
		
		if not AI.Target or not MovableMan:ValidMO(AI.Target) then
			break
		end
		
		if Owner.EquippedItem then
			local Dist = SceneMan:ShortestDistance(Owner.EquippedItem.Pos, AI.Target.Pos, false)
			if Dist.Magnitude < 250 then		--CHANGE THIS: you could change this to shorten the distance before the AI starts attacking
				AI.Ctrl.AnalogAim = SceneMan:ShortestDistance(Owner.EyePos, AI.Target.Pos, false).Normalized
				AI.fire = true
			else
				AI.fire = false
			end
		-- else TODO: periodically look for weapons?
		end
	end
	
	return true
end


-- move around another actor
function MeleeHumanBehaviors.MoveAroundActor(AI, Owner)
	if not Owner.Jetpack or not MovableMan:ValidMO(AI.BlockingMO) then
		AI.teamBlockState = Actor.NOTBLOCKED
		AI.BlockingMO = nil
		return true
	end
	
	local BurstTimer = Timer()
	local refuel = false
	local Dist
	
	BurstTimer:SetSimTimeLimitMS(math.max(SceneMan.GlobalAcc.Y*5, AI.minBurstTime))	-- a burst last until the BurstTimer expire
	AI.jump = true
	
	-- look above the blocking actor
	Dist = SceneMan:ShortestDistance(Owner.Pos, AI.BlockingMO.Pos, false)
	if Dist.X > 0 then
		AI.Ctrl.AnalogAim = Vector(1,0):RadRotate(1.20)
	else
		AI.Ctrl.AnalogAim = Vector(1,0):RadRotate(1.94)
	end
	
	while true do
		if BurstTimer:IsPastSimTimeLimit() then	-- trigger jetpack bursts
			BurstTimer:Reset()
			AI.jump = false
			
			Dist = SceneMan:ShortestDistance(Owner.Pos, AI.BlockingMO.Pos, false)
			if Dist.Y + Owner.Vel.Y * 3 > (Owner.Diameter + AI.BlockingMO.Diameter)*0.67 then
				Owner:SetAimAngle(-0.5)
				
				if math.abs(Dist.X) > math.max(Owner.Diameter, AI.BlockingMO.Diameter)/2 then
					return true
				end
			end
		else
			AI.jump = true
			if Owner.Vel.Y < -9 then
				AI.jump = false
			end
		end
		
		if refuel then
			AI.jump = false
			if Owner.JetTimeLeft > Owner.JetTimeTotal * 0.9 then
				refuel = false
			end
		elseif Owner.JetTimeLeft < Owner.JetTimeTotal * 0.1 then
			refuel = true
		end
		
		coroutine.yield()	-- wait until next frame
		if not MovableMan:IsActor(AI.BlockingMO) then
			AI.teamBlockState = Actor.NOTBLOCKED
			AI.BlockingMO = nil
			return true
		end
	end
	
	return true
end


function MeleeHumanBehaviors.GetAngleToHit(PrjDat, Dist)
	if PrjDat.g == 0 then	-- this projectile is not affected by gravity 
		return Dist.AbsRadAngle
	else	-- compensate for gravity
		local rootSq, muzVelSq
		local D = Dist / FrameMan.PPM	-- convert from pixels to meters
		if PrjDat.drg < 1 then	-- compensate for air resistance
			local rng = D.Magnitude
			local timeToTarget = math.floor((rng / math.max(PrjDat.vel*PrjDat.drg^math.floor(rng/(PrjDat.vel+1)+0.5), PrjDat.thr)) / TimerMan.DeltaTimeSecs)	-- estimate time of flight in frames
			
			if timeToTarget > 1 then
				local muzVel = 0.9*math.max(PrjDat.vel * PrjDat.drg^timeToTarget, PrjDat.thr) + 0.1*PrjDat.vel	-- compensate for velocity reduction during flight
				muzVelSq = muzVel * muzVel
				rootSq = muzVelSq*muzVelSq - PrjDat.g * (PrjDat.g*D.X*D.X + 2*-D.Y*muzVelSq)
			else
				muzVelSq = PrjDat.vsq
				rootSq = PrjDat.vqu - PrjDat.g * (PrjDat.g*D.X*D.X + 2*-D.Y*muzVelSq)
			end
		else
			muzVelSq = PrjDat.vsq
			rootSq = PrjDat.vqu - PrjDat.g * (PrjDat.g*D.X*D.X + 2*-D.Y*muzVelSq)
		end
		
		if rootSq >= 0 then	-- no solution exists if rootSq is below zero
			local ang1 = math.atan2(muzVelSq - math.sqrt(rootSq), PrjDat.g*D.X)
			local ang2 = math.atan2(muzVelSq + math.sqrt(rootSq), PrjDat.g*D.X)
			if ang1 + ang2 > math.pi then	-- both angles in the second or third quadrant
				if ang1 > math.pi or ang2 > math.pi then	-- one or more angle in the third quadrant
					return math.min(ang1, ang2)
				else
					return math.max(ang1, ang2)
				end
			else	-- both angles in the firs quadrant
				return math.min(ang1, ang2)
			end
		end
	end
end

-- open fire on the area around the selected target
function MeleeHumanBehaviors.ShootArea(AI, Owner)
	if not MovableMan:ValidMO(AI.UnseenTarget) or not Owner.FirearmIsReady then
		return true
	end
	
	-- see if we can shoot from the prone position
	local ShootTimer = Timer()
	local aimTime = 50 + RangeRand(100, 300) * AI.aimSpeed
	if not AI.flying and AI.UnseenTarget.Vel.Largest < 12 and MeleeHumanBehaviors.GoProne(AI, Owner, AI.UnseenTarget.Pos, AI.UnseenTarget.ID) then
		aimTime = aimTime + 500
	end
	
	local StartPos = Vector(AI.UnseenTarget.Pos.X, AI.UnseenTarget.Pos.Y)
	
	-- aim at the target in case we can see it when sharp aiming
	Owner:SetAimAngle(SceneMan:ShortestDistance(Owner.EyePos, StartPos, false).AbsRadAngle)
	AI.deviceState = AHuman.AIMING
	
	-- aim for ~160ms
	for _ = 1, 10 do
		coroutine.yield()	-- wait until next frame
	end
	
	if not Owner.FirearmIsReady then
		return true
	end
	
	local AimPoint
	for _ = 1, 5 do	-- try up to five times to find a target area that is resonably close to the target
		AimPoint = StartPos + Vector(RangeRand(-100, 100), RangeRand(-100, 50))
		if AimPoint.X >= SceneMan.SceneWidth then
			AimPoint.X = SceneMan.SceneWidth - AimPoint.X
		elseif AimPoint.X < 0 then
			AimPoint.X = AimPoint.X + SceneMan.SceneWidth
		end
		
		-- check if we can fire at the AimPoint
		local Trace = SceneMan:ShortestDistance(Owner.EyePos, AimPoint, false)
		local rayLenght = SceneMan:CastObstacleRay(Owner.EyePos, Trace, Vector(), Vector(), rte.NoMOID, Owner.IgnoresWhichTeam, rte.grassID, 11)
		if Trace.Magnitude * 0.67 < rayLenght then
			break	-- the AimPoint is close enough to the target, start shooting
		end
		
		coroutine.yield()	-- wait until next frame
	end
	
	if not Owner.FirearmIsReady then
		return true
	end
	
	local aim
	local PrjDat = MeleeHumanBehaviors.GetProjectileData(Owner)
	local Dist = SceneMan:ShortestDistance(Owner.EquippedItem.Pos, AimPoint, false)
	local Weapon = ToHDFirearm(Owner.EquippedItem)
	
	-- uncomment these to get the range of the weapon
	--ConsoleMan:PrintString(Owner.EquippedItem.PresetName .. " range = " .. PrjDat.rng .. " px")
	--ConsoleMan:PrintString("AimPoint range = " .. SceneMan:ShortestDistance(Owner.Pos, AimPoint, false).Magnitude .. " px")
	
	if Dist.Magnitude < PrjDat.rng then
		aim = MeleeHumanBehaviors.GetAngleToHit(PrjDat, Dist)
	else
		return true	-- target out of range
	end
	
	local CheckTargetTimer = Timer()
	local aimError = RangeRand(-0.25, 0.25) * AI.aimSkill
	
	AI.fire = false
	while aim do
		if Owner.FirearmIsReady then
			AI.deviceState = AHuman.AIMING
			AI.Ctrl.AnalogAim = Vector(1,0):RadRotate(aim+aimError+RangeRand(-0.02, 0.02)*AI.aimSkill)
			if ShootTimer:IsPastRealMS(aimTime) then
				if Weapon.FullAuto then
					AI.fire = true
				else
					ShootTimer:Reset()
					aimTime = 120 * AI.aimSkill
					AI.fire = not AI.fire
				end
				
				aimError = aimError * 0.985
			end
		else
			AI.deviceState = AHuman.POINTING
			AI.fire = false
			
			ShootTimer:Reset()
			if Owner.FirearmIsEmpty then
				Owner:ReloadFirearm()
			end
			
			break -- stop this behavior when the mag is empty
		end
		
		coroutine.yield()	-- wait until next frame
		
		if AI.UnseenTarget and CheckTargetTimer:IsPastRealMS(400*AI.aimSkill) then
			if MovableMan:ValidMO(AI.UnseenTarget) and (AI.UnseenTarget.ClassName == "AHuman" or AI.UnseenTarget.ClassName == "ACrab") then
				CheckTargetTimer:Reset()
				if AI.UnseenTarget:GetController() and AI.UnseenTarget:GetController():IsState(Controller.WEAPON_FIRE) then
					-- compare the enemy aim angle with the angle to us
					local AimEnemy = SceneMan:ShortestDistance(AI.UnseenTarget.EyePos, AI.UnseenTarget.ViewPoint, false).Normalized
					local DistNormal = SceneMan:ShortestDistance(AI.UnseenTarget.EyePos, Owner.Pos, false).Normalized
					local dot = DistNormal.X * AimEnemy.X + DistNormal.Y * AimEnemy.Y
					if dot > 0.4 then
						-- this actor is shooting in our direction
						AimPoint = AI.UnseenTarget.Pos + 
							SceneMan:ShortestDistance(AI.UnseenTarget.Pos, AimPoint, false) / 2 +
							Vector(RangeRand(-40, 40)*AI.aimSkill, RangeRand(-40, 40)*AI.aimSkill)
						aimError = RangeRand(-0.15, 0.15) * AI.aimSkill
						
						Dist = SceneMan:ShortestDistance(Owner.EquippedItem.Pos, AimPoint, false)
						if Dist.Magnitude < PrjDat.rng then
							aim = MeleeHumanBehaviors.GetAngleToHit(PrjDat, Dist)
						end
					end
				end
			else
				AI.UnseenTarget = nil
			end
		end
	end
	
	return true
end

-- look at the alarm event
function MeleeHumanBehaviors.FaceAlarm(AI, Owner)
	if AI.AlarmPos then
		local AlarmDist = SceneMan:ShortestDistance(Owner.EyePos, AI.AlarmPos, false)
		AI.AlarmPos = nil
		for _ = 1, math.ceil(200/TimerMan.DeltaTimeMS) do
			AI.deviceState = AHuman.AIMING
			AI.lateralMoveState = Actor.LAT_STILL
			AI.Ctrl.AnalogAim = AlarmDist.Normalized
			coroutine.yield()	-- wait until next frame
		end
	end
	
	return true
end

-- sharp aim at an area where we expect the enemy to be
function MeleeHumanBehaviors.PinArea(AI, Owner)
	if AI.OldTargetPos then
		local AlarmDist = SceneMan:ShortestDistance(Owner.EyePos, AI.OldTargetPos, false)
		for _ = 1, math.ceil(math.random(1000, 3000)/TimerMan.DeltaTimeMS) do
			AI.deviceState = AHuman.AIMING
			AI.lateralMoveState = Actor.LAT_STILL
			AlarmDist:SetXY(AlarmDist.X+RangeRand(-5,5), AlarmDist.Y+RangeRand(-5,5))
			AI.Ctrl.AnalogAim = AlarmDist.Normalized
			coroutine.yield()	-- wait until next frame
		end
	end
	
	return true
end

-- stop the user from inadvertently modifying the storage table
local Proxy = {}
local Mt = {
	__index = MeleeHumanBehaviors,
	__newindex = function(Table, k, v)
		error("The MeleeHumanBehaviors table is read-only.", 2)
	end
}
setmetatable(Proxy, Mt)
MeleeHumanBehaviors = Proxy
